# Copyright (C) 2022 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Create, configure and insert EAO."""
# (c) 2021 ANSYS, Inc. Unauthorized use, distribution, or duplication is prohibited.
#
# Create, configure and insert an Evolutionary Algorithm Optimizer into
# the scenery.

import PyNOA

# Construct a n settings object
ea_settings = PyNOA.NOASettings()

# Configure some global settings
ea_settings.set_defaults()
ea_settings.set_start_pop_size(45)
ea_settings.set_min_gen(10)

# Set and configure the first adaption method
ea_settings.set_adaption_method(PyNOA.Multipoint)
ea_settings.set_number_of_crosspoints(3)
ea_settings.set_crossover_probability(0.3)
ea_settings.set_distribution_parameter(3)

# For hybrid optimization, append another adaption method,
# then use the same functions as before to configure the
# second adaption.
ea_settings.set_hybrid(True)
ea_settings.append_adaption_method(PyNOA.Multipoint)
ea_settings.set_number_of_crosspoints(2)
ea_settings.set_crossover_probability(0.2)
ea_settings.set_distribution_parameter(2)

# Note: to change the settings for the first adaption method afterwards, you
# have to start again with the routine beginning on line 14:
# 1. set the first method and configure it
# 2. append the second method and configure it


# Create the actor
ea = actors.EAActor("My_EA")
# Apply the settings
ea.optimizer_settings = ea_settings
# Add the actor to the scenery
add_actor(ea)



# Copyright (C) 2022 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Module for generation of all server queries."""
import json
from typing import Any, Dict, Optional

QueryArgs = Dict[str, Any]

_ACTOR_INFO = "ACTOR_INFO"
_ACTOR_INTERNAL_VARIABLES = "ACTOR_INTERNAL_VARIABLES"
_ACTOR_PROPERTIES = "ACTOR_PROPERTIES"
_ACTOR_REGISTERED_INPUT_SLOTS = "ACTOR_REGISTERED_INPUT_SLOTS"
_ACTOR_REGISTERED_OUTPUT_SLOTS = "ACTOR_REGISTERED_OUTPUT_SLOTS"
_ACTOR_REGISTERED_PARAMETERS = "ACTOR_REGISTERED_PARAMETERS"
_ACTOR_REGISTERED_RESPONSES = "ACTOR_REGISTERED_RESPONSES"
_ACTOR_STATES = "ACTOR_STATES"
_ACTOR_STATUS_INFO = "ACTOR_STATUS_INFO"
_ACTOR_SUPPORTS = "ACTOR_SUPPORTS"
_AVAILABLE_INPUT_LOCATIONS = "AVAILABLE_INPUT_LOCATIONS"
_AVAILABLE_NODES = "AVAILABLE_NODES"
_AVAILABLE_OUTPUT_LOCATIONS = "AVAILABLE_OUTPUT_LOCATIONS"
_BASIC_PROJECT_INFO = "BASIC_PROJECT_INFO"
_DOE_SIZE = "DOE_SIZE"
_FULL_PROJECT_STATUS_INFO = "FULL_PROJECT_STATUS_INFO"
_FULL_PROJECT_TREE = "FULL_PROJECT_TREE"
_FULL_PROJECT_TREE_WITH_PROPERTIES = "FULL_PROJECT_TREE_WITH_PROPERTIES"
_GET_CRITERIA = "GET_CRITERIA"
_GET_CRITERION = "GET_CRITERION"
_GET_DESIGNS = "GET_DESIGNS"
_HPC_LICENSING_FORWARDED_ENVIRONMENT = "HPC_LICENSING_FORWARDED_ENVIRONMENT"
_INPUT_SLOT_VALUE = "INPUT_SLOT_VALUE"
_OUTPUT_SLOT_VALUE = "OUTPUT_SLOT_VALUE"
_PROJECT_TREE_SYSTEMS = "PROJECT_TREE_SYSTEMS"
_PROJECT_TREE_SYSTEMS_WITH_PROPERTIES = "PROJECT_TREE_SYSTEMS_WITH_PROPERTIES"
_RESULT_DESIGN = "RESULT_DESIGN"
_SERVER_INFO = "SERVER_INFO"
_SERVER_IS_ALIVE = "SERVER_IS_ALIVE"
_SYSTEMS_STATUS_INFO = "SYSTEMS_STATUS_INFO"
_FULL_SUBTREE_STATUS_INFO = "FULL_SUBTREE_STATUS_INFO"


def actor_info(
    uid: str,
    include_log_messages: bool = True,
    include_integrations_registered_locations: bool = True,
    password: Optional[str] = None,
) -> str:
    """Generate JSON string of actor_info query.

    Parameters
    ----------
    uid: str
        Uid entry.
    include_log_messages: bool, optional
        Whether actor log messages are to be included.
    include_integrations_registered_locations: bool, optional
        Whether registered integration locations are to be included.
    password : Optional[str], optional
        Password, by default ``None``.

    Returns
    -------
    str
        JSON string of actor_info query.
    """
    args: QueryArgs = {}
    args["include_log_messages"] = include_log_messages
    args["include_integrations_registered_locations"] = include_integrations_registered_locations
    return _to_json(_gen_query(what=_ACTOR_INFO, uid=uid, args=args, password=password))


def actor_internal_variables(
    uid: str, include_reference_values: bool = True, password: Optional[str] = None
) -> str:
    """Generate JSON string of actor_internal_variables query.

    Parameters
    ----------
    uid: str
        Uid entry.
    include_reference_values: bool, optional
        Whether reference values are to be included.
    password : Optional[str], optional
        Password. Defaults to ``None``.

    Returns
    -------
    str
        JSON string of actor_internal_variables query.
    """
    return _to_json(
        _gen_query(
            what=_ACTOR_INTERNAL_VARIABLES,
            uid=uid,
            args={"include_reference_values": include_reference_values},
            password=password,
        )
    )


def actor_properties(uid: str, password: Optional[str] = None) -> str:
    """Generate JSON string of actor_properties query.

    Parameters
    ----------
    uid: str
        Uid entry.
    password : Optional[str], optional
        Password, by default ``None``.

    Returns
    -------
    str
        JSON string of actor_properties query.
    """
    return _to_json(_gen_query(what=_ACTOR_PROPERTIES, uid=uid, password=password))


def actor_registered_input_slots(
    uid: str, include_reference_values: bool = True, password: Optional[str] = None
) -> str:
    """Generate JSON string of actor_registered_input_slots query.

    Parameters
    ----------
    uid: str
        Uid entry.
    include_reference_values: bool, optional
        Whether reference values are to be included.
    password : Optional[str], optional
        Password. Defaults to ``None``.

    Returns
    -------
    str
        JSON string of actor_registered_input_slots query.
    """
    return _to_json(
        _gen_query(
            what=_ACTOR_REGISTERED_INPUT_SLOTS,
            uid=uid,
            args={"include_reference_values": include_reference_values},
            password=password,
        )
    )


def actor_registered_output_slots(
    uid: str, include_reference_values: bool = True, password: Optional[str] = None
) -> str:
    """Generate JSON string of actor_registered_output_slots query.

    Parameters
    ----------
    uid: str
        Uid entry.
    include_reference_values: bool, optional
        Whether reference values are to be included.
    password : Optional[str], optional
        Password. Defaults to ``None``.

    Returns
    -------
    str
        JSON string of actor_registered_output_slots query.
    """
    return _to_json(
        _gen_query(
            what=_ACTOR_REGISTERED_OUTPUT_SLOTS,
            uid=uid,
            args={"include_reference_values": include_reference_values},
            password=password,
        )
    )


def actor_registered_parameters(
    uid: str, include_reference_values: bool = True, password: Optional[str] = None
) -> str:
    """Generate JSON string of actor_registered_parameters query.

    Parameters
    ----------
    uid: str
        Uid entry.
    include_reference_values: bool, optional
        Whether reference values are to be included.
    password : Optional[str], optional
        Password. Defaults to ``None``.

    Returns
    -------
    str
        JSON string of actor_registered_parameters query.
    """
    return _to_json(
        _gen_query(
            what=_ACTOR_REGISTERED_PARAMETERS,
            uid=uid,
            args={"include_reference_values": include_reference_values},
            password=password,
        )
    )


def actor_registered_responses(
    uid: str, include_reference_values: bool = True, password: Optional[str] = None
) -> str:
    """Generate JSON string of actor_registered_responses query.

    Parameters
    ----------
    uid: str
        Uid entry.
    include_reference_values: bool, optional
        Whether reference values are to be included.
    password : Optional[str], optional
        Password. Defaults to ``None``.

    Returns
    -------
    str
        JSON string of actor_registered_responses query.
    """
    return _to_json(
        _gen_query(
            what=_ACTOR_REGISTERED_RESPONSES,
            uid=uid,
            args={"include_reference_values": include_reference_values},
            password=password,
        )
    )


def actor_states(
    uid: str,
    include_state_info: bool = False,
    password: Optional[str] = None,
) -> str:
    """Generate JSON string of actor_states query.

    Parameters
    ----------
    uid: str
        Uid entry.
    include_state_info: bool
        Include additional info for each state. Otherwise, only state IDs are returned.
    password : Optional[str], optional
        Password. Defaults to ``None``.

    Returns
    -------
    str
        JSON string of actor_states query.
    """
    args: QueryArgs = {}
    args["include_state_info"] = include_state_info
    return _to_json(_gen_query(what=_ACTOR_STATES, uid=uid, args=args, password=password))


def actor_status_info(
    uid: str,
    hid: str,
    include_designs: bool = True,
    include_design_values: bool = True,
    include_non_scalar_design_values: bool = False,
    include_algorithm_info: bool = False,
    password: Optional[str] = None,
) -> str:
    """Generate JSON string of actor_status_info query.

    Parameters
    ----------
    uid: str
        Uid entry.
    hid: str
        Hid entry.
    include_designs: bool
        Include (result) designs in status info response.
    include_design_values: bool
        Include values in (result) designs.
    include_non_scalar_design_values: bool
        Include non scalar values in (result) designs.
    include_algorithm_info: bool
        Include algorithm result info in status info response.
    password : Optional[str], optional
        Password. Defaults to ``None``.

    Returns
    -------
    str
        JSON string of actor_status_info query.
    """
    args: QueryArgs = {}
    args["include_designs"] = include_designs
    args["include_design_values"] = include_design_values
    args["include_non_scalar_design_values"] = include_non_scalar_design_values
    args["include_algorithm_info"] = include_algorithm_info
    return _to_json(
        _gen_query(what=_ACTOR_STATUS_INFO, uid=uid, args=args, hid=hid, password=password)
    )


def actor_supports(uid: str, feature_name: str, password: Optional[str] = None) -> str:
    """Generate JSON string of actor_supports query.

    Parameters
    ----------
    uid: str
        Uid entry.
    feature_name: str
        Name of requested feature.
    password : Optional[str], optional
        Password, by default ``None``.

    Returns
    -------
    str
        JSON string of actor_supports query.
    """
    return _to_json(
        _gen_query(what=_ACTOR_SUPPORTS, uid=uid, args={"feature": feature_name}, password=password)
    )


def available_input_locations(uid: str, password: Optional[str] = None) -> str:
    """Generate JSON string of available_input_locations query.

    Parameters
    ----------
    uid: str
        Uid entry.
    password : Optional[str], optional
        Password. Defaults to ``None``.

    Returns
    -------
    str
        JSON string of available_input_locations query.
    """
    return _to_json(
        _gen_query(
            what=_AVAILABLE_INPUT_LOCATIONS,
            uid=uid,
            password=password,
        )
    )


def available_nodes(password: Optional[str] = None) -> str:
    """Generate JSON string of available nodes query.

    Parameters
    ----------
    password : Optional[str], optional
        Password, by default ``None``.

    Returns
    -------
    str
        JSON string of available nodes query.
    """
    return _to_json(_gen_query(what=_AVAILABLE_NODES, password=password))


def available_output_locations(uid: str, password: Optional[str] = None) -> str:
    """Generate JSON string of available_output_locations query.

    Parameters
    ----------
    uid: str
        Uid entry.
    password : Optional[str], optional
        Password. Defaults to ``None``.

    Returns
    -------
    str
        JSON string of available_output_locations query.
    """
    return _to_json(
        _gen_query(
            what=_AVAILABLE_OUTPUT_LOCATIONS,
            uid=uid,
            password=password,
        )
    )


def basic_project_info(password: Optional[str] = None) -> str:
    """Generate JSON string of basic_project_info query.

    Parameters
    ----------
    password : Optional[str], optional
        Password, by default ``None``.

    Returns
    -------
    str
        JSON string of basic_project_info query.
    """
    return _to_json(_gen_query(what=_BASIC_PROJECT_INFO, password=password))


def doe_size(
    uid: str, sampling_type: str, num_discrete_levels: int, password: Optional[str] = None
) -> str:
    """Generate JSON string of doe_size query.

    Parameters
    ----------
    uid: str
        Uid entry.
    sampling_type: str
        Sampling type.
    num_discrete_levels: int
        Number of discrete levels.
    password : Optional[str], optional
        Password. Defaults to ``None``.

    Returns
    -------
    str
        JSON string of doe_size query.
    """
    args: QueryArgs = {}
    args["sampling_type"] = sampling_type
    args["num_discrete_levels"] = num_discrete_levels
    return _to_json(
        _gen_query(
            what=_DOE_SIZE,
            uid=uid,
            args=args,
            password=password,
        )
    )


def full_project_status_info(
    include_designs: bool = True,
    include_design_values: bool = True,
    include_non_scalar_design_values: bool = False,
    include_algorithm_info: bool = False,
    include_log_messages: bool = True,
    include_integrations_registered_locations: bool = True,
    password: Optional[str] = None,
) -> str:
    """Generate JSON string of full_project_status_info query.

    Parameters
    ----------
    include_designs: bool
        Include (result) designs in status info response.
    include_design_values: bool
        Include values in (result) designs.
    include_non_scalar_design_values: bool
        Include non scalar values in (result) designs.
    include_algorithm_info: bool
        Include algorithm result info in status info response.
    include_log_messages: bool, optional
        Whether actor log messages are to be included.
    include_integrations_registered_locations: bool, optional
        Whether registered integration locations are to be included.
    password : Optional[str], optional
        Password. Defaults to ``None``.

    Returns
    -------
    str
        JSON string of full_project_status_info query.
    """
    args: QueryArgs = {}
    args["include_designs"] = include_designs
    args["include_design_values"] = include_design_values
    args["include_non_scalar_design_values"] = include_non_scalar_design_values
    args["include_algorithm_info"] = include_algorithm_info
    args["include_log_messages"] = include_log_messages
    args["include_integrations_registered_locations"] = include_integrations_registered_locations
    return _to_json(_gen_query(what=_FULL_PROJECT_STATUS_INFO, args=args, password=password))


def full_project_tree(password: Optional[str] = None) -> str:
    """Generate JSON string of full_project_tree query.

    Parameters
    ----------
    password : Optional[str], optional
        Password, by default ``None``.

    Returns
    -------
    str
        JSON string of full_project_tree query.
    """
    return _to_json(_gen_query(what=_FULL_PROJECT_TREE, password=password))


def full_project_tree_with_properties(password: Optional[str] = None) -> str:
    """Generate JSON string of full_project_tree_with_properties query.

    Parameters
    ----------
    password : Optional[str], optional
        Password, by default ``None``.

    Returns
    -------
    str
        JSON string of full_project_tree_with_properties query.
    """
    return _to_json(_gen_query(what=_FULL_PROJECT_TREE_WITH_PROPERTIES, password=password))


def full_subtree_status_info(
    uid: str,
    include_designs: bool = True,
    include_design_values: bool = True,
    include_non_scalar_design_values: bool = False,
    include_algorithm_info: bool = False,
    include_log_messages: bool = True,
    include_integrations_registered_locations: bool = True,
    password: Optional[str] = None,
) -> str:
    """Generate JSON string of full_subtree_status_info query.

    Parameters
    ----------
    uid: str
        Uid entry.
    include_designs: bool
        Include (result) designs in status info response.
    include_design_values: bool
        Include values in (result) designs.
    include_non_scalar_design_values: bool
        Include non scalar values in (result) designs.
    include_algorithm_info: bool
        Include algorithm result info in status info response.
    include_log_messages: bool, optional
        Whether actor log messages are to be included.
    include_integrations_registered_locations: bool, optional
        Whether registered integration locations are to be included.
    password : Optional[str], optional
        Password. Defaults to ``None``.

    Returns
    -------
    str
        JSON string of systems_status_info query.
    """
    args: QueryArgs = {}
    args["include_designs"] = include_designs
    args["include_design_values"] = include_design_values
    args["include_non_scalar_design_values"] = include_non_scalar_design_values
    args["include_algorithm_info"] = include_algorithm_info
    args["include_log_messages"] = include_log_messages
    args["include_integrations_registered_locations"] = include_integrations_registered_locations
    return _to_json(
        _gen_query(what=_FULL_SUBTREE_STATUS_INFO, uid=uid, args=args, password=password)
    )


def get_criteria(uid: str, password: Optional[str] = None) -> str:
    """Generate JSON string of get_criteria query.

    Parameters
    ----------
    uid: str
        Uid entry.
    password : Optional[str], optional
        Password. Defaults to ``None``.

    Returns
    -------
    str
        JSON string of get_criteria query.
    """
    return _to_json(_gen_query(what=_GET_CRITERIA, uid=uid, password=password))


def get_criterion(uid: str, name: str, password: Optional[str] = None) -> str:
    """Generate JSON string of get_criterion query.

    Parameters
    ----------
    uid: str
        Uid entry.
    name: str
        Criterion name.
    password : Optional[str], optional
        Password. Defaults to ``None``.

    Returns
    -------
    str
        JSON string of get_criterion query.
    """
    return _to_json(
        _gen_query(what=_GET_CRITERION, uid=uid, args={"name": name}, password=password)
    )


def get_designs(uid: str, password: Optional[str] = None) -> str:
    """Generate JSON string of get_designs query.

    Parameters
    ----------
    uid: str
        Uid entry.
    password : Optional[str], optional
        Password. Defaults to ``None``.

    Returns
    -------
    str
        JSON string of get_designs query.
    """
    return _to_json(_gen_query(what=_GET_DESIGNS, uid=uid, password=password))


def hpc_licensing_forwarded_environment(uid: str, password: Optional[str] = None) -> str:
    """Generate JSON string of hpc_licensing_forwarded_environment query.

    Parameters
    ----------
    uid: str
        Uid entry.
    password : Optional[str], optional
        Password, by default ``None``.

    Returns
    -------
    str
        JSON string of hpc_licensing_forwarded_environment query.
    """
    return _to_json(
        _gen_query(what=_HPC_LICENSING_FORWARDED_ENVIRONMENT, uid=uid, password=password)
    )


def input_slot_value(uid: str, hid: str, slot_name: str, password: Optional[str] = None) -> str:
    """Generate JSON string of input_slot_value query.

    Parameters
    ----------
    uid: str
        Uid entry.
    hid: str
        Hid entry.
    slot_name: str
        Slot name entry.
    password : Optional[str], optional
        Password, by default ``None``.

    Returns
    -------
    str
        JSON string of input_slot_value query.
    """
    return _to_json(
        _gen_query(what=_INPUT_SLOT_VALUE, uid=uid, hid=hid, slot_name=slot_name, password=password)
    )


def output_slot_value(uid: str, hid: str, slot_name: str, password: Optional[str] = None) -> str:
    """Generate JSON string of output_slot_value query.

    Parameters
    ----------
    uid: str
        Uid entry.
    hid: str
        Hid entry.
    slot_name: str
        Slot name entry.
    password : Optional[str], optional
        Password, by default ``None``.

    Returns
    -------
    str
        JSON string of output_slot_value query.
    """
    return _to_json(
        _gen_query(
            what=_OUTPUT_SLOT_VALUE, uid=uid, hid=hid, slot_name=slot_name, password=password
        )
    )


def project_tree_systems(password: Optional[str] = None) -> str:
    """Generate JSON string of project_tree_systems query.

    Parameters
    ----------
    password : Optional[str], optional
        Password, by default ``None``.

    Returns
    -------
    str
        JSON string of project_tree_systems query.
    """
    return _to_json(_gen_query(what=_PROJECT_TREE_SYSTEMS, password=password))


def project_tree_systems_with_properties(password: Optional[str] = None) -> str:
    """Generate JSON string of project_tree_with_properties query.

    Parameters
    ----------
    password : Optional[str], optional
        Password, by default ``None``.

    Returns
    -------
    str
        JSON string of project_tree_with_properties query.
    """
    return _to_json(_gen_query(what=_PROJECT_TREE_SYSTEMS_WITH_PROPERTIES, password=password))


def result_design(
    uid: str,
    design_id: str,
    password: Optional[str] = None,
) -> str:
    """Generate JSON string of result_design query.

    Parameters
    ----------
    uid: str
        Uid entry.
    design_id: str
        Design ID entry.
    password : Optional[str], optional
        Password. Defaults to ``None``.

    Returns
    -------
    str
        JSON string of result_design query.
    """
    return _to_json(
        _gen_query(what=_RESULT_DESIGN, uid=uid, design_id=design_id, password=password)
    )


def server_info(password: Optional[str] = None) -> str:
    """Generate JSON string of server_info query.

    Parameters
    ----------
    password : Optional[str], optional
        Password, by default ``None``.

    Returns
    -------
    str
        JSON string of server_info query.
    """
    return _to_json(_gen_query(what=_SERVER_INFO, password=password))


def server_is_alive(password: Optional[str] = None) -> str:
    """Generate JSON string of server_is_alive query.

    Parameters
    ----------
    password : Optional[str], optional
        Password, by default ``None``.

    Returns
    -------
    str
        JSON string of server_is_alive query.
    """
    return _to_json(_gen_query(what=_SERVER_IS_ALIVE, password=password))


def systems_status_info(
    include_designs: bool = True,
    include_design_values: bool = True,
    include_non_scalar_design_values: bool = False,
    include_algorithm_info: bool = False,
    include_log_messages: bool = True,
    include_integrations_registered_locations: bool = True,
    password: Optional[str] = None,
) -> str:
    """Generate JSON string of systems_status_info query.

    Parameters
    ----------
    include_designs: bool
        Include (result) designs in status info response.
    include_design_values: bool
        Include values in (result) designs.
    include_non_scalar_design_values: bool
        Include non scalar values in (result) designs.
    include_algorithm_info: bool
        Include algorithm result info in status info response.
    include_log_messages: bool, optional
        Whether actor log messages are to be included.
    include_integrations_registered_locations: bool, optional
        Whether registered integration locations are to be included.
    password : Optional[str], optional
        Password. Defaults to ``None``.

    Returns
    -------
    str
        JSON string of systems_status_info query.
    """
    args: QueryArgs = {}
    args["include_designs"] = include_designs
    args["include_design_values"] = include_design_values
    args["include_non_scalar_design_values"] = include_non_scalar_design_values
    args["include_algorithm_info"] = include_algorithm_info
    args["include_log_messages"] = include_log_messages
    args["include_integrations_registered_locations"] = include_integrations_registered_locations
    return _to_json(_gen_query(what=_SYSTEMS_STATUS_INFO, args=args, password=password))


def _gen_query(
    what: str,
    password: Optional[str],
    uid: Optional[str] = None,
    hid: Optional[str] = None,
    args: Optional[QueryArgs] = None,
    slot_name: Optional[str] = None,
    design_id: Optional[str] = None,
) -> Dict:
    """Generate query in desired format.

    Parameters
    ----------
    what : str
        Command type.
    password: Optional[str], optional
        Password. Defaults to ``None``.
    uid : Optional[str], optional
        Uid entry. Defaults to ``None``.
    hid: Optional[str], optional
        Hid entry. Defaults to ``None``.
    args: Optional[QueryArgs], optional
        Dictionary of features, e.g. "feature": "FEATURE_NAME".
    slot_name: Optional[str], optional
        Slot name. Defaults to ``None``.
    design_id: Optional[str], optional
        Design ID. Defaults to ``None``.

    Returns
    -------
    Dict
        Dictionary with specified query inputs.
    """
    query: Dict[str, Any] = {}
    query["What"] = what
    if uid is not None:
        query["uid"] = uid
    if hid is not None:
        query["hid"] = hid
    if args is not None:
        query["args"] = args
    if slot_name is not None:
        query["slot_name"] = slot_name
    if design_id is not None:
        query["design_id"] = design_id
    if password is not None:
        query["Password"] = password
    return query


def _to_json(dict: Dict) -> str:
    """Convert dictionary to JSON.

    Parameters
    ----------
    dict : Dict
        Dictionary with specified query inputs.

    Returns
    -------
    str
        JSON string.
    """
    return json.dumps(dict, sort_keys=True, ensure_ascii=False)



# Copyright (C) 2022 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Module for generation of all server commands."""
import json
from typing import Any, Dict, Iterable, Optional, Sequence, Union

CommandArgs = Dict[str, Any]

_APPLY_WIZARD = "APPLY_WIZARD"
_ADD_CRITERION = "ADD_CRITERION"
_CLOSE = "CLOSE"
_CONNECT_NODES = "CONNECT_NODES"
_CREATE_INPUT_SLOT = "CREATE_INPUT_SLOT"
_CREATE_NODE = "CREATE_NODE"
_CREATE_OUTPUT_SLOT = "CREATE_OUTPUT_SLOT"
_CREATE_START_DESIGNS = "CREATE_START_DESIGNS"
_DISCONNECT_SLOT = "DISCONNECT_SLOT"
_EVALUATE_DESIGN = "EVALUATE_DESIGN"
_EXPORT_DESIGNS = "EXPORT_DESIGNS"
_FINALIZE = "FINALIZE"
_LINK_REGISTERED_FILE = "LINK_REGISTERED_FILE"
_LOAD = "LOAD"
_NEW = "NEW"
_OPEN = "OPEN"
_PAUSE = "PAUSE"
_REEVALUATE_STATE = "REEVALUATE_STATE"
_REFRESH_LISTENER_REGISTRATION = "REFRESH_LISTENER_REGISTRATION"
_REGISTER_FILE = "REGISTER_FILE"
_REGISTER_LISTENER = "REGISTER_LISTENER"
_REGISTER_LOCATION_AS_INPUT_SLOT = "REGISTER_LOCATION_AS_INPUT_SLOT"
_REGISTER_LOCATION_AS_INTERNAL_VARIABLE = "REGISTER_LOCATION_AS_INTERNAL_VARIABLE"
_REGISTER_LOCATION_AS_OUTPUT_SLOT = "REGISTER_LOCATION_AS_OUTPUT_SLOT"
_REGISTER_LOCATION_AS_PARAMETER = "REGISTER_LOCATION_AS_PARAMETER"
_REGISTER_LOCATIONS_AS_PARAMETER = "REGISTER_LOCATIONS_AS_PARAMETER"
_REGISTER_LOCATION_AS_RESPONSE = "REGISTER_LOCATION_AS_RESPONSE"
_REGISTER_LOCATIONS_AS_RESPONSE = "REGISTER_LOCATIONS_AS_RESPONSE"
_REMOVE_CRITERIA = "REMOVE_CRITERIA"
_REMOVE_CRITERION = "REMOVE_CRITERION"
_REMOVE_NODE = "REMOVE_NODE"
_RE_REGISTER_LOCATIONS_AS_PARAMETER = "RE_REGISTER_LOCATIONS_AS_PARAMETER"
_RE_REGISTER_LOCATIONS_AS_RESPONSE = "RE_REGISTER_LOCATIONS_AS_RESPONSE"
_RESET = "RESET"
_RESTART = "RESTART"
_RESUME = "RESUME"
_RUN_PYTHON_SCRIPT = "RUN_PYTHON_SCRIPT"
_RUN_REGISTERED_FILES_ACTIONS = "RUN_REGISTERED_FILES_ACTIONS"
_SAVE = "SAVE"
_SAVE_AS = "SAVE_AS"
_SAVE_COPY = "SAVE_COPY"
_SET_ACTOR_PROPERTY = "SET_ACTOR_PROPERTY"
_SET_ACTOR_SETTING = "SET_ACTOR_SETTING"
_SET_ACTOR_STATE_PROPERTY = "SET_ACTOR_STATE_PROPERTY"
_SET_CRITERION_PROPERTY = "SET_CRITERION_PROPERTY"
_SET_DESIGNS = "SET_DESIGNS"
_SET_PLACEHOLDER_VALUE = "SET_PLACEHOLDER_VALUE"
_SET_PROJECT_SETTING = "SET_PROJECT_SETTING"
_SET_REGISTERED_FILE_VALUE = "SET_REGISTERED_FILE_VALUE"
_SET_START_DESIGNS = "SET_START_DESIGNS"
_SET_SUCCEEDED_STATE = "SET_SUCCEEDED_STATE"
_SHOW_DIALOG = "SHOW_DIALOG"
_SHOW_NODE_DIALOG = "SHOW_NODE_DIALOG"
_SHUTDOWN = "SHUTDOWN"
_SHUTDOWN_WHEN_FINISHED = "SHUTDOWN_WHEN_FINISHED"
_START = "START"
_STOP = "STOP"
_STOP_GENTLY = "STOP_GENTLY"
_SUBSCRIBE_FOR_PUSH_NOTIFICATIONS = "SUBSCRIBE_FOR_PUSH_NOTIFICATIONS"
_UNLINK_REGISTERED_FILE = "UNLINK_REGISTERED_FILE"
_UNREGISTER_FILE = "UNREGISTER_FILE"
_UNREGISTER_LISTENER = "UNREGISTER_LISTENER"
_WRITE_MONITORING_DATABASE = "WRITE_MONITORING_DATABASE"
_builtin = "builtin"


def add_criterion(
    actor_uid: str,
    criterion_type: str,
    expression: str,
    name: str,
    limit: str = "",
    password: Optional[str] = None,
) -> str:
    """Generate JSON string of add_criterion command.

    Parameters
    ----------
    actor_uid: str
        Unique identifying actor of the object.
    criterion_type: str
        Type of the criterion. Supported values are:
        ['ignore', 'min', 'max', 'lessequal', 'equal', 'greaterequal',
        'lesslimitstate', 'greaterlimitstate']
    expression: str
        Expression to be evaluated.
    name: str
        Criterion name.
    limit: str, optional
        Limit expression. By default `""`.
    password : Optional[str], optional
        Password. Defaults to ``None``.

    Returns
    -------
    str
        JSON string of add_criterion command.
    """
    args: CommandArgs = {}
    args["criterion_type"] = criterion_type
    args["expression"] = expression
    args["name"] = name
    args["limit"] = limit if limit is not None else ""

    return _to_json(
        _gen_server_command(
            command=_ADD_CRITERION, actor_uid=actor_uid, args=args, password=password
        )
    )


def apply_wizard(
    actor_uid: str,
    type_: str,
    use_existing_system: Optional[bool] = None,
    usage_mode: Optional[str] = None,
    parent_hwnd: Optional[str] = None,
    password: Optional[str] = None,
) -> str:
    """Generate JSON string of apply_wizzard command.

    Parameters
    ----------
    actor_uid: str
        Unique identifying actor of the object.
    type_: str
        Node or system type, supported values are:
        ["solver", "sensitivity", "optimization", "robustness", "reevaluation"].
    use_existing_system: Optional[bool], optional
        True or false. Defaults to ``None``.
    usage_mode: Optional[str], optional
        Usage mode, e.g. "EXPERT". Defaults to ``None``.
    parent_hwnd: Optional[str], optional
        Parent hwnd. Defaults to ``None``.
    password : Optional[str], optional
        Password. Defaults to ``None``.

    Returns
    -------
    str
        JSON string of apply_wizzard command.
    """
    args: CommandArgs = {}
    supported_values = ["solver", "sensitivity", "optimization", "robustness", "reevaluation"]
    if type_ in supported_values:
        args["type"] = type_
    else:
        raise TypeError(
            f"Unsuppored value of ``type_``: {type_}, supported values are: {supported_values}"
        )
    if use_existing_system is not None:
        args["use_existing_system"] = use_existing_system
    if usage_mode is not None:
        args["usage_mode"] = usage_mode
    # TODO: and/or:  The request must contain the actor_uid entry as well as the args entry,
    # containing the argument type and optionally the arguments usage_mode,
    # use_existing_system and/or parent_hwnd."
    if parent_hwnd is not None:
        args["parent_hwnd"] = parent_hwnd

    return _to_json(
        _gen_server_command(
            command=_APPLY_WIZARD, actor_uid=actor_uid, args=args, password=password
        )
    )


def close(password: Optional[str] = None) -> str:
    """Generate JSON string of close command.

    Parameters
    ----------
    password : Optional[str], optional
        Password, by default ``None``.

    Returns
    -------
    str
        JSON string of close command.
    """
    return _to_json(_gen_server_command(command=_CLOSE, password=password))


def connect_nodes(
    from_actor_uid: str,
    from_slot: str,
    to_actor_uid: str,
    to_slot: str,
    password: Optional[str] = None,
) -> str:
    """Generate JSON string of connect_nodes command.

    Parameters
    ----------
    from_actor_uid: str
        Uid of connection source.
    from_slot: str
        Slot of connection source.
    to_actor_uid: str
        Uid of connection target.
    to_slot: str
        Slot of connection target.
    password : Optional[str], optional
        Password, by default ``None``.

    Returns
    -------
    str
        JSON string of connect_nodes command.
    """
    args: CommandArgs = {}
    args["from_actor_uid"] = from_actor_uid
    args["from_slot"] = from_slot
    args["to_actor_uid"] = to_actor_uid
    args["to_slot"] = to_slot

    return _to_json(_gen_server_command(command=_CONNECT_NODES, args=args, password=password))


def create_input_slot(
    actor_uid: str, slot_name: str, type_hint: Optional[str] = None, password: Optional[str] = None
) -> str:
    """Generate JSON string of create_input_slot command.

    Parameters
    ----------
    actor_uid: str
        Actor uid entry.
    slot_name: str
        Name of slot.
    type_hint: Optional[str], optional
        Type of hint. Defaults to ``None``.
    password : Optional[str], optional
        Password. Defaults to ``None``.

    Returns
    -------
    str
        JSON string of create_input_slot command.
    """
    args: CommandArgs = {}
    args["slot_name"] = slot_name
    if type_hint is not None:
        args["type_hint"] = type_hint

    return _to_json(
        _gen_server_command(
            command=_CREATE_INPUT_SLOT, actor_uid=actor_uid, args=args, password=password
        )
    )


def create_node(
    type_: str,
    name: Optional[str] = None,
    algorithm_type: Optional[str] = None,
    integration_type: Optional[str] = None,
    mop_node_type: Optional[str] = None,
    node_type: Optional[str] = None,
    parent_uid: Optional[str] = None,
    design_flow: Optional[str] = None,
    password: Optional[str] = None,
) -> str:
    """Generate JSON string of create_node command.

    Parameters
    ----------
    type_: str
        Type.
    name: Optional[str], optional
        Name, by default ``None``.
    algorithm_type : Optional[str], optional
        Algorithm type, e. g. 'algorithm_plugin', by default ``None``.
    integration_type : Optional[str], optional
        Integration type, e. g. 'integration_plugin', by default ``None``.
    mop_node_type : Optional[str], optional
        MOP node type, e. g. 'python_based_mop_node_plugin', by default ``None``.
    node_type: Optional[str], optional
        Node type, e. g. 'python_based_node_plugin`, by default ``None``.
    parent_uid: Optional[str], optional
        Parent uid entry, by default ``None``.
    design_flow: Optional[str], optional
        Design flow, optional values are ["RECEIVE", "SEND", "RECEIVE_SEND"], by default ``None``.
    password : Optional[str], optional
        Password, by default ``None``.

    Returns
    -------
    str
        JSON string of create_node command.
    """
    args: CommandArgs = {}
    args["type"] = type_
    if name:
        args["name"] = name
    if algorithm_type:
        args["algorithm_type"] = algorithm_type
    if integration_type:
        args["integration_type"] = integration_type
    if mop_node_type:
        args["mop_node_type"] = mop_node_type
    if node_type:
        args["node_type"] = node_type
    if parent_uid is not None:
        args["parent_uid"] = parent_uid
    if design_flow not in [None, "NONE"]:
        args["design_flow"] = design_flow

    return _to_json(_gen_server_command(command=_CREATE_NODE, args=args, password=password))


def create_output_slot(
    actor_uid: str, slot_name: str, type_hint: Optional[str] = None, password: Optional[str] = None
) -> str:
    """Generate JSON string of create_output_slot command.

    Parameters
    ----------
    actor_uid: str
        Actor uid entry.
    slot_name: str
        Name of the slot.
    type_hint: Optional[str], optional
        Type of the hint, by default ``None``.
    password : Optional[str], optional
        Password, by default ``None``.

    Returns
    -------
    str
        JSON string of create_output_slot command.
    """
    args: CommandArgs = {}
    args["slot_name"] = slot_name
    if type_hint is not None:
        args["type_hint"] = type_hint
    return _to_json(
        _gen_server_command(
            command=_CREATE_OUTPUT_SLOT, actor_uid=actor_uid, args=args, password=password
        )
    )


def create_start_designs(
    actor_uid: str,
    sampling_type: str,
    number_of_levels: Optional[int] = None,
    number_of_samples: Optional[int] = None,
    password: Optional[str] = None,
) -> str:
    """Generate JSON string of create_start_design command.

    Parameters
    ----------
    actor_uid: str
        Actor uid entry.
    sampling_type: str
        Sampling type, e.g. "fullfactorial".
    number_of_levels: Optional[int], optional
        Number of levels, by default ``None``.
    number_of_samples: Optional[int], optional
        Number of samples, by default ``None``.
    password : Optional[str], optional
        Password, by default ``None``.

    Returns
    -------
    str
        JSON string of create_start_design command.
    """
    args: CommandArgs = {}
    args["sampling_type"] = sampling_type
    if (number_of_levels is not None) and (number_of_samples is not None):
        raise TypeError("Please specify either ``number_of_levels`` or ``number of samples``.")
    elif number_of_levels is not None:
        args["number_of_levels"] = number_of_levels
    elif number_of_samples is not None:
        args["number_of_samples"] = number_of_samples

    return _to_json(
        _gen_server_command(
            command=_CREATE_START_DESIGNS, actor_uid=actor_uid, args=args, password=password
        )
    )


def disconnect_slot(
    actor_uid: str, slot_name: str, direction: str, password: Optional[str] = None
) -> str:
    """Generate JSON string of disconnect_slot command.

    Parameters
    ----------
    actor_uid: str
        Actor uid entry.
    slot_name: str
        Name of the slot.
    direction: str
        Direction.
    password : Optional[str], optional
        Password, by default ``None``.

    Returns
    -------
    str
        JSON string of disconnect_slot command.
    """
    args: CommandArgs = {}
    args["slot_name"] = slot_name
    args["direction"] = direction
    return _to_json(
        _gen_server_command(
            command=_DISCONNECT_SLOT, actor_uid=actor_uid, args=args, password=password
        )
    )


def evaluate_design(parameters: Dict, password: Optional[str] = None) -> str:
    """Generate JSON string of evaluate_design command.

    Parameters
    ----------
    parameters: Dict
        Dictionary of parameters.
    password : Optional[str], optional
        Password, by default ``None``.

    Returns
    -------
    str
        JSON string of evaluate_design command.
    """
    args: CommandArgs = {}
    args["parameters"] = parameters
    return _to_json(_gen_server_command(command=_EVALUATE_DESIGN, args=args, password=password))


def export_designs(
    actor_uid: str,
    path: str,
    format: Optional[str] = None,
    csv_separator: Optional[str] = None,
    password: Optional[str] = None,
) -> str:
    """Generate JSON string of export_designs command.

    Parameters
    ----------
    actor_uid: str
        Actor uid entry.
    path: str
       Path.
    format: Optional[str], optional
        Type_hint format, by default ``None``.
    csv_separator: Optional[str], optional
        CSV separator, by default ``None``.
    password : Optional[str], optional
        Password, by default ``None``.

    Returns
    -------
    str
        JSON string of export_designs command.
    """
    args: CommandArgs = {}
    args["path"] = path
    if format is not None:
        args["format"] = format
    if csv_separator is not None:
        args["csv_separator"] = csv_separator

    return _to_json(
        _gen_server_command(
            command=_EXPORT_DESIGNS, actor_uid=actor_uid, args=args, password=password
        )
    )


def finalize(actor_uid: str, password: Optional[str] = None) -> str:
    """Generate JSON string of finalize command.

    Parameters
    ----------
    actor_uid: str
        Actor uid entry.
    password : Optional[str], optional
        Password, by default ``None``.

    Returns
    -------
    str
        JSON string of finalize command.
    """
    return _to_json(_gen_server_command(command=_FINALIZE, actor_uid=actor_uid, password=password))


def link_registered_file(actor_uid: str, uid: str, password: Optional[str] = None) -> str:
    """Generate JSON string of link_registered_file command.

    Parameters
    ----------
    actor_uid: str
        Actor uid entry.
    uid: str
        Uid entry.
    password : Optional[str], optional
        Password, by default ``None``.

    Returns
    -------
    str
        JSON string of link_registered_file command.
    """
    args: CommandArgs = {}
    args["uid"] = uid
    return _to_json(
        _gen_server_command(
            command=_LINK_REGISTERED_FILE, actor_uid=actor_uid, args=args, password=password
        )
    )


def load(actor_uid: str, args: Optional[CommandArgs] = None, password: Optional[str] = None) -> str:
    """Generate JSON string of ``load`` command.

    Parameters
    ----------
    actor_uid: str
        Actor uid entry.
    args: Optional[CommandArgs], optional
        Dictionary with additional arguments, by default ``None``.
    password : Optional[str], optional
        Password, by default ``None``.

    Returns
    -------
    str
        JSON string of ``load`` command.
    """
    return _to_json(
        _gen_server_command(command=_LOAD, args=args, actor_uid=actor_uid, password=password)
    )


def new(password: Optional[str] = None) -> str:
    """Generate JSON string of ``new`` command.

    Parameters
    ----------
    password : Optional[str], optional
        Password, by default ``None``.

    Returns
    -------
    str
        JSON string of ``new`` command.
    """
    return _to_json(_gen_server_command(command=_NEW, password=password))


def open(
    path: str,
    do_force: bool,
    do_restore: bool,
    do_reset: bool,
    password: Optional[str] = None,
    project_properties_file: Optional[str] = None,
) -> str:
    """Generate JSON string of ``open`` command.

    Parameters
    ----------
    path: str
        path.
    do_force: bool
        True/False.
    do_restore: bool
        True/False.
    do_reset: bool
        True/False.
    password : Optional[str], optional
        Password, by default ``None``.
    project_properties_file : Optional[str], optional
        Project properties file to import, by default ``None``.

    Returns
    -------
    str
        JSON string of ``open`` command.
    """
    args: CommandArgs = {}
    args["path"] = path
    args["do_force"] = do_force
    args["do_restore"] = do_restore
    args["do_reset"] = do_reset
    if project_properties_file is not None:
        args["project_properties_file"] = project_properties_file

    return _to_json(_gen_server_command(command=_OPEN, args=args, password=password))


def pause(password: Optional[str] = None) -> str:
    """Generate JSON string of pause command.

    Parameters
    ----------
    password : Optional[str], optional
        Password, by default ``None``.

    Returns
    -------
    str
        JSON string of pause command.
    """
    return _to_json(_gen_server_command(command=_PAUSE, password=password))


def reevaluate_state(
    actor_uid: str, hid: Optional[str] = None, password: Optional[str] = None
) -> str:
    """Generate JSON string of reevaluate_state command.

    Parameters
    ----------
    actor_uid: str
        Actor uid entry.
    hid: Optional[str]
        Hid entry. by default ``None``.
    password : Optional[str], optional
        Password, by default ``None``.

    Returns
    -------
    str
        JSON string of reevaluate_state command.
    """
    return _to_json(
        _gen_server_command(
            command=_REEVALUATE_STATE, actor_uid=actor_uid, hid=hid, password=password
        )
    )


def refresh_listener_registration(uid: str, password: Optional[str] = None) -> str:
    """Generate JSON string of refresh_listener_registration command.

    Parameters
    ----------
    uid: str
        Uid entry.
    password : Optional[str], optional
        Password, by default ``None``.

    Returns
    -------
    str
        JSON string of refresh_listener_registration command.
    """
    args: CommandArgs = {}
    args["uid"] = uid
    return _to_json(
        _gen_server_command(command=_REFRESH_LISTENER_REGISTRATION, args=args, password=password)
    )


def register_file(
    uid: Optional[str] = None,
    ident: Optional[str] = None,
    local_location: Optional[Dict] = None,
    action: Optional[str] = None,
    password: Optional[str] = None,
) -> str:
    """Generate JSON string of register_file command.

    Parameters
    ----------
    uid: Optional[str], optional
        Uid entry, by default ``None``.
    ident: Optional[str], optional
        Ident, by default ``None``.
    local_location: Optional[Dict], optional
        Dictionary specifying location, e. g. { "split_path": {"head": "",
        "tail": "C:/samples_path/result.txt" }, "base_path": "",
        "base_path_mode": "ABSOLUTE_PATH" }, by default ``None``.
    action: Optional[str], optional
        Action, e. g. "Send", by default ``None``.
    password : Optional[str], optional
        Password, by default ``None``.

    Returns
    -------
    str
        JSON string of register_file command.
    """
    # TODO: "and so on: specify all possibilities"
    args: CommandArgs = {}
    if uid is not None:
        args["uid"] = uid
    if ident is not None:
        args["ident"] = ident
    if local_location is not None:
        args["local_location"] = local_location
    if action is not None:
        args["action"] = action

    return _to_json(_gen_server_command(command=_REGISTER_FILE, args=args, password=password))


def register_listener(
    id: Optional[str] = None,
    host: Optional[str] = None,
    port: Optional[int] = None,
    timeout: Optional[int] = None,
    notifications: Optional[Iterable[str]] = None,
    password: Optional[str] = None,
    listener_uid: Optional[str] = None,
) -> str:
    """Generate JSON string of register_listener command.

    Parameters
    ----------
    id: Optional[str], optional
        Id of the local listener, by default ``None``.
    host: Optional[str], optional
        IP of the TCP listener, by default ``None``.
    port: Optional[int], optional
        Port of the TCP listener, by default ``None``.
    time_out: Optional[int], optional
        Unregister policy timeout in ms, default 60000 ms, by default ``None``.
    notifications: Optional[Sequence], optional
        Notifications, by default ``None``.
    password : Optional[str], optional
        Password, by default ``None``.
    listener_uid : Optional[str], optional
        Listener UID, by default ``None``.

    Returns
    -------
    str
        JSON string of register_listener command.
    """
    args: CommandArgs = {}
    if (id is not None) and ((host is not None) or (port is not None)):
        raise TypeError("Please specify either ``id`` or (``host`` and ``port``).")
    elif (id is None) and ((host is None) or (port is None)):
        raise TypeError("Please specify either ``id`` or (``host`` and ``port``).")
    elif id is not None:
        args["id"] = id
    else:
        args["host"] = host
        args["port"] = port

    if timeout is not None:
        args["timeout"] = timeout
    if notifications is not None:
        args["notifications"] = notifications
    if listener_uid is not None:
        args["uid"] = listener_uid

    return _to_json(_gen_server_command(command=_REGISTER_LISTENER, args=args, password=password))


def register_location_as_input_slot(
    actor_uid: str,
    location: dict,
    name: Optional[str] = None,
    reference_value: Optional[Any] = None,
    password: Optional[str] = None,
) -> str:
    """Generate JSON string of ``register_location_as_input_slot`` command.

    Parameters
    ----------
    actor_uid: str
        Actor uid entry.
    location: dict
        Dictionary with location properties.
    name: Optional[str], optional
        Input slot name.
    reference_value: Optional[Any], optional
        Input slot reference value.
    password : Optional[str], optional
        Password, by default ``None``.

    Returns
    -------
    str
        JSON string of ``register_location_as_input_slot`` command.
    """
    args: CommandArgs = {
        "location": location,
    }
    if name is not None:
        args["name"] = name
    if reference_value is not None:
        args["reference_value"] = reference_value

    return _to_json(
        _gen_server_command(
            command=_REGISTER_LOCATION_AS_INPUT_SLOT,
            actor_uid=actor_uid,
            args=args,
            password=password,
        )
    )


def register_location_as_internal_variable(
    actor_uid: str,
    location: dict,
    name: Optional[str] = None,
    reference_value: Optional[Any] = None,
    password: Optional[str] = None,
) -> str:
    """Generate JSON string of ``register_location_as_internal_variable`` command.

    Parameters
    ----------
    actor_uid: str
        Actor uid entry.
    location: dict
        Dictionary with location properties.
    name: Optional[str], optional
        Variable name.
    reference_value: Optional[Any], optional
        Variable reference value.
    password : Optional[str], optional
        Password, by default ``None``.

    Returns
    -------
    str
        JSON string of ``register_location_as_internal_variable`` command.
    """
    args: CommandArgs = {
        "location": location,
    }
    if name is not None:
        args["name"] = name
    if reference_value is not None:
        args["reference_value"] = reference_value

    return _to_json(
        _gen_server_command(
            command=_REGISTER_LOCATION_AS_INTERNAL_VARIABLE,
            actor_uid=actor_uid,
            args=args,
            password=password,
        )
    )


def register_location_as_output_slot(
    actor_uid: str,
    location: dict,
    name: Optional[str] = None,
    reference_value: Optional[Any] = None,
    password: Optional[str] = None,
) -> str:
    """Generate JSON string of ``register_location_as_output_slot`` command.

    Parameters
    ----------
    actor_uid: str
        Actor uid entry.
    location: dict
        Dictionary with location properties.
    name: Optional[str], optional
        Output slot name.
    reference_value: Optional[Any], optional
        Output slot reference value.
    password : Optional[str], optional
        Password, by default ``None``.

    Returns
    -------
    str
        JSON string of ``register_location_as_output_slot`` command.
    """
    args: CommandArgs = {
        "location": location,
    }
    if name is not None:
        args["name"] = name
    if reference_value is not None:
        args["reference_value"] = reference_value

    return _to_json(
        _gen_server_command(
            command=_REGISTER_LOCATION_AS_OUTPUT_SLOT,
            actor_uid=actor_uid,
            args=args,
            password=password,
        )
    )


def register_location_as_parameter(
    actor_uid: str,
    location: dict,
    name: Optional[str] = None,
    reference_value: Optional[Any] = None,
    password: Optional[str] = None,
) -> str:
    """Generate JSON string of ``register_location_as_parameter`` command.

    Parameters
    ----------
    actor_uid: str
        Actor uid entry.
    location: dict
        Dictionary with location properties.
    name: Optional[str], optional
        Parameter name.
    reference_value: Optional[Any], optional
        Parameter reference value.
    password : Optional[str], optional
        Password, by default ``None``.

    Returns
    -------
    str
        JSON string of ``register_location_as_parameter`` command.
    """
    args: CommandArgs = {
        "location": location,
    }
    if name is not None:
        args["name"] = name
    if reference_value is not None:
        args["reference_value"] = reference_value

    return _to_json(
        _gen_server_command(
            command=_REGISTER_LOCATION_AS_PARAMETER,
            actor_uid=actor_uid,
            args=args,
            password=password,
        )
    )


def register_locations_as_parameter(actor_uid: str, password: Optional[str] = None) -> str:
    """Generate JSON string of ``register_locations_as_parameter`` command.

    Parameters
    ----------
    actor_uid: str
        Actor uid entry.
    password : Optional[str], optional
        Password, by default ``None``.

    Returns
    -------
    str
        JSON string of ``register_locations_as_parameter`` command.
    """
    return _to_json(
        _gen_server_command(
            command=_REGISTER_LOCATIONS_AS_PARAMETER, actor_uid=actor_uid, password=password
        )
    )


def register_location_as_response(
    actor_uid: str,
    location: dict,
    name: Optional[str] = None,
    reference_value: Optional[Any] = None,
    password: Optional[str] = None,
) -> str:
    """Generate JSON string of ``register_location_as_response`` command.

    Parameters
    ----------
    actor_uid: str
        Actor uid entry.
    location: dict
        Dictionary with location properties.
    name: Optional[str], optional
        Response name.
    reference_value: Optional[Any], optional
        Response reference value.
    password : Optional[str], optional
        Password, by default ``None``.

    Returns
    -------
    str
        JSON string of ``register_location_as_response`` command.
    """
    args: CommandArgs = {
        "location": location,
    }
    if name is not None:
        args["name"] = name
    if reference_value is not None:
        args["reference_value"] = reference_value

    return _to_json(
        _gen_server_command(
            command=_REGISTER_LOCATION_AS_RESPONSE,
            actor_uid=actor_uid,
            args=args,
            password=password,
        )
    )


def register_locations_as_response(actor_uid: str, password: Optional[str] = None) -> str:
    """Generate JSON string of ``register_locations_as_response`` command.

    Parameters
    ----------
    actor_uid: str
        Actor uid entry.
    password : Optional[str], optional
        Password, by default ``None``.

    Returns
    -------
    str
        JSON string of ``register_locations_as_response`` command.
    """
    return _to_json(
        _gen_server_command(
            command=_REGISTER_LOCATIONS_AS_RESPONSE, actor_uid=actor_uid, password=password
        )
    )


def remove_criteria(actor_uid: str, password: Optional[str] = None) -> str:
    """Generate JSON string of ``remove_criteria`` command.

    Parameters
    ----------
    actor_uid: str
        Actor uid entry.
    password : Optional[str], optional
        Password. Defaults to ``None``.

    Returns
    -------
    str
        JSON string of ``remove_criteria`` command.
    """
    return _to_json(
        _gen_server_command(command=_REMOVE_CRITERIA, actor_uid=actor_uid, password=password)
    )


def remove_criterion(actor_uid: str, name: str, password: Optional[str] = None) -> str:
    """Generate JSON string of ``remove_criterion`` command.

    Parameters
    ----------
    actor_uid: str
        Actor uid entry.
    name: str
        Criterion name.
    password : Optional[str], optional
        Password. Defaults to ``None``.

    Returns
    -------
    str
        JSON string of ``remove_criterion`` command.
    """
    return _to_json(
        _gen_server_command(
            command=_REMOVE_CRITERION, args={"name": name}, actor_uid=actor_uid, password=password
        )
    )


def remove_node(actor_uid: str, password: Optional[str] = None) -> str:
    """Generate JSON string of ``remove node`` command.

    Parameters
    ----------
    actor_uid: str
        Actor uid entry.
    password : Optional[str], optional
        Password, by default ``None``.

    Returns
    -------
    str
        JSON string of ``remove_node`` command.
    """
    return _to_json(
        _gen_server_command(command=_REMOVE_NODE, actor_uid=actor_uid, password=password)
    )


def re_register_locations_as_parameter(actor_uid: str, password: Optional[str] = None) -> str:
    """Generate JSON string of ``re_register_locations_as_parameter`` command.

    Parameters
    ----------
    actor_uid: str
        Actor uid entry.
    password : Optional[str], optional
        Password, by default ``None``.

    Returns
    -------
    str
        JSON string of ``re_register_locations_as_parameter`` command.
    """
    return _to_json(
        _gen_server_command(
            command=_RE_REGISTER_LOCATIONS_AS_PARAMETER, actor_uid=actor_uid, password=password
        )
    )


def re_register_locations_as_response(actor_uid: str, password: Optional[str] = None) -> str:
    """Generate JSON string of ``re_register_locations_as_response`` command.

    Parameters
    ----------
    actor_uid: str
        Actor uid entry.
    password : Optional[str], optional
        Password, by default ``None``.

    Returns
    -------
    str
        JSON string of ``re_register_locations_as_response`` command.
    """
    return _to_json(
        _gen_server_command(
            command=_RE_REGISTER_LOCATIONS_AS_RESPONSE, actor_uid=actor_uid, password=password
        )
    )


def reset(
    actor_uid: Optional[str] = None, hid: Optional[str] = None, password: Optional[str] = None
) -> str:
    """Generate JSON string of ``reset`` command.

    Parameters
    ----------
    actor_uid: Optional[str], optional
        Actor uid entry. A Hierarchical ID (hid) is required. By default ``None``.
    hid: Optional[str], optional
        Hid entry. The actor uid is required. By default ``None``.
    password : Optional[str], optional
        Password, by default ``None``.

    Returns
    -------
    str
        JSON string of ``reset`` command.
    """
    if actor_uid and hid is None:
        raise ValueError("The Hierarchical ID (hid) is required.")
    elif actor_uid is None and hid:
        raise ValueError("The actor uid is required.")
    return _to_json(
        _gen_server_command(command=_RESET, actor_uid=actor_uid, hid=hid, password=password)
    )


def restart(
    actor_uid: Optional[str] = None, hid: Optional[str] = None, password: Optional[str] = None
) -> str:
    """Generate JSON string of ``restart`` command.

    Parameters
    ----------
    actor_uid: Optional[str], optional
        Actor uid entry. A Hierarchical ID (hid) is required. By default ``None``.
    hid: Optional[str], optional
        Hid entry. The actor uid is required. By default ``None``.
    password : Optional[str], optional
        Password, by default ``None``.

    Returns
    -------
    str
        JSON string of ``restart`` command.
    """
    if actor_uid and hid is None:
        raise ValueError("The Hierarchical ID (hid) is required.")
    elif actor_uid is None and hid:
        raise ValueError("The actor uid is required.")
    return _to_json(
        _gen_server_command(command=_RESTART, actor_uid=actor_uid, hid=hid, password=password)
    )


def resume(password: Optional[str] = None) -> str:
    """Generate JSON string of ``resume`` command.

    Parameters
    ----------
    password : Optional[str], optional
        Password, by default ``None``.

    Returns
    -------
    str
        JSON string of ``resume`` command.
    """
    return _to_json(_gen_server_command(command=_RESUME, password=password))


def run_python_script(
    script: str,
    args_: Optional[list] = None,
    password: Optional[str] = None,
) -> str:
    """Generate JSON string of register_listener command.

    Parameters
    ----------
    script: str
        Path of the script.
    args: Optional[list], optional
        Script arguments, by default ``None``.
    password : Optional[str], optional
        Password, by default ``None``.

    Returns
    -------
    str
        JSON string of register_listener command.
    """
    args: CommandArgs = {}
    args["script"] = script
    if args_ is not None:
        args["args"] = args_

    return _to_json(_gen_server_command(command=_RUN_PYTHON_SCRIPT, args=args, password=password))


def run_registered_files_actions(
    uid: Optional[str] = None,
    password: Optional[str] = None,
) -> str:
    """Generate JSON string of ``run registered files actions`` command.

    Parameters
    ----------
    uid: Optional[str], optional
        Uid entry, by default ``None``.
    password : Optional[str], optional
        Password, by default ``None``.

    Returns
    -------
    str
        JSON string of ``run registered files actions`` command.
    """
    args: Optional[CommandArgs] = None
    if uid is not None:
        args = {}
        args["uid"] = uid

    return _to_json(
        _gen_server_command(command=_RUN_REGISTERED_FILES_ACTIONS, args=args, password=password)
    )


def save(password: Optional[str] = None) -> str:
    """Generate JSON string of ``save`` command.

    Parameters
    ----------
    password : Optional[str], optional
        Password, by default ``None``.

    Returns
    -------
    str
        JSON string of ``save`` command.
    """
    return _to_json(_gen_server_command(command=_SAVE, password=password))


def save_as(
    path: str, do_force: bool, do_restore: bool, do_reset: bool, password: Optional[str] = None
) -> str:
    """Generate JSON string of ``save_as`` command.

    Parameters
    ----------
    path: str
        path.
    do_force: bool
        Whether to force opening of project even if (non-critical) errors occur.
            Non-critical errors include:
            - Timestamp of (auto) save point newer than project timestamp
            - Project (file) incomplete
    do_restore: bool
        Whether to restore project from last (auto) save point (if present).
    do_reset: bool
        Whether to reset project after load.
    password : Optional[str], optional
        Password, by default ``None``.

    Returns
    -------
    str
        JSON string of ``save_as`` command.
    """
    args: CommandArgs = {}
    args["path"] = path
    args["do_force"] = do_force
    args["do_restore"] = do_restore
    args["do_reset"] = do_reset

    return _to_json(_gen_server_command(command=_SAVE_AS, args=args, password=password))


def save_copy(path: str, password: Optional[str] = None) -> str:
    """Generate JSON string of ``save_copy`` command.

    Parameters
    ----------
    path: str
        path.
    password : Optional[str], optional
        Password, by default ``None``.

    Returns
    -------
    str
        JSON string of ``save_copy`` command.
    """
    args: CommandArgs = {}
    args["path"] = path

    return _to_json(_gen_server_command(command=_SAVE_COPY, args=args, password=password))


def set_actor_property(
    actor_uid: str, name: str, value: str, password: Optional[str] = None
) -> str:
    """Generate JSON string of ``set_actor_property`` command.

    Parameters
    ----------
    actor_uid: str
        Actor uid entry.
    name: str
        Property name.
    value: str
        Value.
    password : Optional[str], optional
        Password, by default ``None``.

    Returns
    -------
    str
        JSON string of ``set_actor_property`` command.
    """
    args: CommandArgs = {}
    args["name"] = name
    args["value"] = value

    return _to_json(
        _gen_server_command(
            command=_SET_ACTOR_PROPERTY, actor_uid=actor_uid, args=args, password=password
        )
    )


def set_actor_setting(actor_uid: str, name: str, value: str, password: Optional[str] = None) -> str:
    """Generate JSON string of ``set actor setting`` command.

    Parameters
    ----------
    actor_uid: str
        Actor uid entry.
    name: str
        Property name.
    value: str
        Value.
    password : Optional[str], optional
        Password, by default ``None``.

    Returns
    -------
    str
        JSON string of ``set actor setting`` command.
    """
    args: CommandArgs = {}
    args["name"] = name
    args["value"] = value

    return _to_json(
        _gen_server_command(
            command=_SET_ACTOR_SETTING, actor_uid=actor_uid, args=args, password=password
        )
    )


def set_actor_state_property(
    actor_uid: str, name: str, value: str, password: Optional[str] = None
) -> str:
    """Generate JSON string of ``set actor state property`` command.

    Parameters
    ----------
    actor_uid: str
        Actor uid entry.
    name: str
        Property name.
    value: str
        Value.
    password : Optional[str], optional
        Password, by default ``None``.

    Returns
    -------
    str
        JSON string of ``set actor state property`` command.
    """
    args: CommandArgs = {}
    args["name"] = name
    args["value"] = value

    return _to_json(
        _gen_server_command(
            command=_SET_ACTOR_STATE_PROPERTY, actor_uid=actor_uid, args=args, password=password
        )
    )


def set_criterion_property(
    actor_uid: str, criterion_name: str, name: str, value: Any, password: Optional[str] = None
) -> str:
    """Generate JSON string of ``set_criterion_property`` command.

    Parameters
    ----------
    actor_uid: str
        Actor uid entry.
    criterion_name: str
        Name of the criterion.
    name: str
        Name of the property.
    value: Any
        Value of the property.
    password : Optional[str], optional
        Password. Defaults to ``None``.

    Returns
    -------
    str
        JSON string of ``set_criterion_property`` command.
    """
    args: CommandArgs = {}
    args["criterion_name"] = criterion_name
    args["name"] = name
    args["value"] = value

    return _to_json(
        _gen_server_command(
            command=_SET_CRITERION_PROPERTY, actor_uid=actor_uid, args=args, password=password
        )
    )


def set_designs(actor_uid: str, designs: Iterable[dict], password: Optional[str] = None) -> str:
    """Generate JSON string of ``set_designs`` command.

    Parameters
    ----------
    actor_uid: str
        Actor uid entry.
    designs: Iterable[dict]
        Iterable of calculated designs.
    password : Optional[str], optional
        Password. Defaults to ``None``.

    Returns
    -------
    str
        JSON string of ``set_designs`` command.
    """
    args: CommandArgs = {}
    args["designs"] = designs

    return _to_json(
        _gen_server_command(command=_SET_DESIGNS, actor_uid=actor_uid, args=args, password=password)
    )


def set_placeholder_value(name: str, value: str, password: Optional[str] = None) -> str:
    """Generate JSON string of ``set placeholder value`` command.

    Parameters
    ----------
    name: str
        Property name.
    value: str
        Value.
    password : Optional[str], optional
        Password, by default ``None``.

    Returns
    -------
    str
        JSON string of ``set placeholder value`` command.
    """
    args: CommandArgs = {}
    args["name"] = name
    args["value"] = value

    return _to_json(
        _gen_server_command(command=_SET_PLACEHOLDER_VALUE, args=args, password=password)
    )


def set_project_setting(name: str, value: str, password: Optional[str] = None) -> str:
    """Generate JSON string of ``set project settings`` command.

    Parameters
    ----------
    name: str
        Property name.
    value: str
        Value.
    password : Optional[str], optional
        Password, by default ``None``.

    Returns
    -------
    str
        JSON string of ``set project settings`` command.
    """
    args: CommandArgs = {}
    args["name"] = name
    args["value"] = value

    return _to_json(_gen_server_command(command=_SET_PROJECT_SETTING, args=args, password=password))


def set_registered_file_value(
    uid: str, name: str, value: Union[str, Dict], password: Optional[str] = None
) -> str:
    """Generate JSON string of ``set registered file value`` command.

    Parameters
    ----------
    uid: str
        Uid entry.
    name: str
        Property name.
    value: Union[str, Dict]
        Value.
    password : Optional[str], optional
        Password, by default ``None``.

    Returns
    -------
    str
        JSON string of ``set registered file value`` command.
    """
    args: CommandArgs = {}
    args["uid"] = uid
    args["name"] = name
    args["value"] = value

    return _to_json(
        _gen_server_command(command=_SET_REGISTERED_FILE_VALUE, args=args, password=password)
    )


def set_start_designs(actor_uid: str, start_designs: Dict, password: Optional[str] = None) -> str:
    """Generate JSON string of ``set start designs`` command.

    Parameters
    ----------
    actor_uid: str
        Actor uid entry.
    start_designs: Dict
        Dictionary of settings.
    password : Optional[str], optional
        Password, by default ``None``.

    Returns
    -------
    str
        JSON string of ``set start designs`` command.
    """
    # TODO: function to create start_settings?
    args: CommandArgs = {}
    args["start_designs"] = start_designs

    return _to_json(
        _gen_server_command(
            command=_SET_START_DESIGNS, actor_uid=actor_uid, args=args, password=password
        )
    )


def set_succeeded_state(
    actor_uid: str, hid: Optional[str] = None, password: Optional[str] = None
) -> str:
    """Generate JSON string of ``set succeeded state`` command.

    Parameters
    ----------
    actor_uid: str
        Actor uid entry.
    hid: Optional[str], optional
        Hid entry, by default ``None``.
    password : Optional[str], optional
        Password, by default ``None``.

    Returns
    -------
    str
        JSON string of ``set succeeded state`` command.
    """
    return _to_json(
        _gen_server_command(
            command=_SET_SUCCEEDED_STATE, actor_uid=actor_uid, hid=hid, password=password
        )
    )


def show_dialog(
    type_: str,
    usage_mode: Optional[str] = None,
    parent_hwnd: Optional[str] = None,
    password: Optional[str] = None,
) -> str:
    """Generate JSON string of ``show dialog`` command.

    Parameters
    ----------
    type_: str
        Argument type,
        for application dialogs: [ "about", "help", "settings", "plugin" ]
        for project dialogs: [ "project_settings", "project_overview", "license_management",
        "registered_files", "purge", "load_from", "save_to" ].
    usage_mode: Optional[str], optional
        Usage mode. e.g. "EXPERT", by default ``None``.
    parent_hwnd: Optional[str], optional
        Parent hwnd, by default ``None``.
    password : Optional[str], optional
        Password, by default ``None``.

    Returns
    -------
    str
        JSON string of ``show dialog`` command.
    """
    application_dialogs = ["about", "help", "settings", "plugin"]
    project_dialogs = [
        "project_settings",
        "project_overview",
        "license_management",
        "registered_files",
        "purge",
        "load_from",
        "save_to",
    ]

    if not ((type_ in application_dialogs) or (type_ in project_dialogs)):
        raise TypeError(
            f"Unsuppored value of ``type_``: {type_}, supported values for application dialogs: "
            "{application_dialogs} and for project dialogs: {project_dialogs}"
        )
    args: CommandArgs = {}
    args["type"] = type_
    if usage_mode is not None:
        args["usage_mode"] = usage_mode
    if parent_hwnd is not None:
        args["parent_hwnd"] = parent_hwnd

    return _to_json(_gen_server_command(command=_SHOW_DIALOG, args=args, password=password))


def show_node_dialog(
    actor_uid: str,
    blocking: Optional[bool] = None,
    type_: Optional[str] = None,
    usage_mode: Optional[str] = None,
    parent_hwnd: Optional[str] = None,
    password: Optional[str] = None,
) -> str:
    """Generate JSON string of ``show node dialog`` command.

    Parameters
    ----------
    actor_uid: str
        Actor uid entry.
    blocking: Optional[bool], optional
        True/False, by default ``None``.
    type_: Optional[str], optional
        Type, e.g. "help", by default ``None``.
    usage_mode: Optional[str], optional
        Usage mode. e.g. "EXPERT", by default ``None``.
    parent_hwnd: Optional[str], optional
        Parent hwnd, by default ``None``.
    password : Optional[str], optional
        Password, by default ``None``.

    Returns
    -------
    str
        JSON string of ``show node dialog`` command.
    """
    args: Optional[CommandArgs] = None
    if not (
        (blocking is None) and (type_ is None) and (usage_mode is None) and (parent_hwnd is None)
    ):
        args = {}
        if blocking is not None:
            args["blocking"] = blocking
        if type_ is not None:
            args["type"] = type_
        if usage_mode is not None:
            args["usage_mode"] = usage_mode
        if parent_hwnd is not None:
            args["parent_hwnd"] = parent_hwnd

    return _to_json(
        _gen_server_command(
            command=_SHOW_NODE_DIALOG, actor_uid=actor_uid, args=args, password=password
        )
    )


def shutdown(force: bool = False, password: Optional[str] = None) -> str:
    """Generate JSON string of ``shutdown`` command.

    Parameters
    ----------
    password : Optional[str], optional
        Password, by default ``None``.

    Returns
    -------
    str
        JSON string of ``shutdown`` command.
    """
    args: CommandArgs = {}
    args["force"] = force

    return _to_json(_gen_server_command(command=_SHUTDOWN, args=args, password=password))


def shutdown_when_finished(password: Optional[str] = None) -> str:
    """Generate JSON string of ``shutdown_when_finished`` command.

    Parameters
    ----------
    password : Optional[str], optional
        Password, by default ``None``.

    Returns
    -------
    str
        JSON string of ``shutdown_when_finished`` command.
    """
    return _to_json(_gen_server_command(command=_SHUTDOWN_WHEN_FINISHED, password=password))


def start(
    actor_uid: Optional[str] = None, hid: Optional[str] = None, password: Optional[str] = None
) -> str:
    """Generate JSON string of ``start`` command.

    Parameters
    ----------
    actor_uid: Optional[str], optional
        Actor uid entry. A Hierarchical ID (hid) is required. By default ``None``.
    hid: Optional[str], optional
        Hid entry. The actor uid is required. By default ``None``.
    password : Optional[str], optional
        Password, by default ``None``.

    Returns
    -------
    str
        JSON string of ``start`` command.
    """
    if actor_uid and hid is None:
        raise ValueError("The Hierarchical ID (hid) is required.")
    elif actor_uid is None and hid:
        raise ValueError("The actor uid is required.")

    return _to_json(
        _gen_server_command(command=_START, actor_uid=actor_uid, hid=hid, password=password)
    )


def stop(
    actor_uid: Optional[str] = None, hid: Optional[str] = None, password: Optional[str] = None
) -> str:
    """Generate JSON string of ``stop`` command.

    Parameters
    ----------
    actor_uid: Optional[str], optional
        Actor uid entry. A Hierarchical ID (hid) is required. By default ``None``.
    hid: Optional[str], optional
        Hid entry. The actor uid is required. By default ``None``.
    password : Optional[str], optional
        Password, by default ``None``.

    Returns
    -------
    str
        JSON string of ``stop`` command.
    """
    if actor_uid and hid is None:
        raise ValueError("The Hierarchical ID (hid) is required.")
    elif actor_uid is None and hid:
        raise ValueError("The actor uid is required.")

    return _to_json(
        _gen_server_command(command=_STOP, actor_uid=actor_uid, hid=hid, password=password)
    )


def stop_gently(
    actor_uid: Optional[str] = None, hid: Optional[str] = None, password: Optional[str] = None
) -> str:
    """Generate JSON string of ``stop_gently`` command.

    Parameters
    ----------
    actor_uid: Optional[str], optional
        Actor uid entry. A Hierarchical ID (hid) is required. By default ``None``.
    hid: Optional[str], optional
        Hid entry. The actor uid is required. By default ``None``.
    password : Optional[str], optional
        Password, by default ``None``.

    Returns
    -------
    str
        JSON string of ``stop_gently`` command.
    """
    if actor_uid and hid is None:
        raise ValueError("The Hierarchical ID (hid) is required.")
    elif actor_uid is None and hid:
        raise ValueError("The actor uid is required.")
    return _to_json(
        _gen_server_command(command=_STOP_GENTLY, actor_uid=actor_uid, hid=hid, password=password)
    )


def subscribe_for_push_notifications(
    uid: str,
    notifications: Sequence,
    node_types: Optional[Sequence] = None,
    password: Optional[str] = None,
) -> str:
    """Generate JSON string of ``subscribe_for_push_notifications`` command.

    Parameters
    ----------
    uid: str
        Uid entry.
    notifications: Sequence
        Either ["ALL"] or Sequence picked from below options:
            Server: [ "SERVER_UP", "SERVER_DOWN" ] (always be sent by default).
            Logging: [ "LOG_INFO", "LOG_WARNING", "LOG_ERROR", "LOG_DEBUG" ].
            Project: [ "EXECUTION_STARTED", "PROCESSING_STARTED", "EXECUTION_FINISHED",
            "NOTHING_PROCESSED", "CHECK_FAILED", "EXEC_FAILED" ].
            Nodes: [ "ACTOR_STATE_CHANGED", "ACTOR_ACTIVE_CHANGED", "ACTOR_NAME_CHANGED",
            ACTOR_CONTENTS_CHANGED", "ACTOR_DATA_CHANGED" ].
    node_types: Optional[Sequence], optional
       Node types, e.g. ["Sensitivity", "AnsysWorkbench"]. By default ``None``.
    password : Optional[str], optional
        Password, by default ``None``.

    Returns
    -------
    str
        JSON string of ``subscribe_for_push_notifications`` command.
    """
    server = ["SERVER_UP", "SERVER_DOWN"]
    logging = ["LOG_INFO", "LOG_WARNING", "LOG_ERROR", "LOG_DEBUG"]
    project = [
        "EXECUTION_STARTED",
        "PROCESSING_STARTED",
        "EXECUTION_FINISHED",
        "NOTHING_PROCESSED",
        "CHECK_FAILED",
        "EXEC_FAILED",
    ]
    nodes = [
        "ACTOR_STATE_CHANGED",
        "ACTOR_ACTIVE_CHANGED",
        "ACTOR_NAME_CHANGED",
        "ACTOR_CONTENTS_CHANGED",
        "ACTOR_DATA_CHANGED",
    ]
    all_opts = set(server + logging + project + nodes)

    args: CommandArgs = {}
    args["uid"] = uid

    if notifications == ["ALL"]:
        args["notifications"] = notifications
    else:
        if not all(item in all_opts for item in notifications):
            invalid_items = list(sorted(all_opts - set(notifications)))
            raise TypeError(
                f"Unsuppored values of ``notifications``: {invalid_items}, "
                "supported options are: \n"
                f"server: {server},\n"
                f"logging: {logging},\n"
                f"project: {project},\n"
                f"nodes: {nodes}"
            )
        args["notifications"] = notifications

    if node_types is not None:
        args["node_types"] = node_types

    return _to_json(
        _gen_server_command(command=_SUBSCRIBE_FOR_PUSH_NOTIFICATIONS, args=args, password=password)
    )


def unlink_registered_file(actor_uid: str, uid: str, password: Optional[str] = None) -> str:
    """Generate JSON string of ``unlink registered file`` command.

    Parameters
    ----------
    actor_uid: str
        Actor uid entry.
    uid: str
        Uid entry.
    password : Optional[str], optional
        Password, by default ``None``.

    Returns
    -------
    str
        JSON string of ``unlink registered file`` command.
    """
    args: CommandArgs = {}
    args["uid"] = uid

    return _to_json(
        _gen_server_command(
            command=_UNLINK_REGISTERED_FILE, actor_uid=actor_uid, args=args, password=password
        )
    )


def unregister_file(uid: str, password: Optional[str] = None) -> str:
    """Generate JSON string of ``unregister file`` command.

    Parameters
    ----------
    uid: str
        Uid entry.
    password : Optional[str], optional
        Password, by default ``None``.

    Returns
    -------
    str
        JSON string of ``unregister file`` command.
    """
    args: CommandArgs = {}
    args["uid"] = uid

    return _to_json(_gen_server_command(command=_UNREGISTER_FILE, args=args, password=password))


def unregister_listener(uid: str, password: Optional[str] = None) -> str:
    """Generate JSON string of ``unregister listener`` command.

    Parameters
    ----------
    uid: str
        Uid entry.
    password : Optional[str], optional
        Password, by default ``None``.

    Returns
    -------
    str
        JSON string of ``unregister listener`` command.
    """
    args: CommandArgs = {}
    args["uid"] = uid

    return _to_json(_gen_server_command(command=_UNREGISTER_LISTENER, args=args, password=password))


def write_monitoring_database(
    actor_uid: str,
    path: Optional[str] = None,
    hid: Optional[str] = None,
    password: Optional[str] = None,
) -> str:
    """Generate JSON string of ``write monitoring database`` command.

    Parameters
    ----------
    actor_uid: str
        Actor uid entry.
    path: Optional[str], optional
        Path, by default ``None``.
    hid: Optional[str], optional
        Hid entry, by default ``None``.
    password : Optional[str], optional
        Password, by default ``None``.

    Returns
    -------
    str
        JSON string of ``write monitoring database`` command.
    """
    args: Optional[CommandArgs] = None
    if (path is not None) or (hid is not None):
        args = {}
        if path is not None:
            args["path"] = path
        if hid is not None:
            args["hid"] = hid

    return _to_json(
        _gen_server_command(
            command=_WRITE_MONITORING_DATABASE, actor_uid=actor_uid, args=args, password=password
        )
    )


def _gen_server_command(
    command: str,
    password: Optional[str] = None,
    args: Optional[CommandArgs] = None,
    actor_uid: Optional[str] = None,
    hid: Optional[str] = None,
) -> Dict:
    """Generate server command.

    Parameters
    ----------
    command : str
        Command type.
    password : Optional[str]
        Password, by default ``None``.
    args : Optional[CommandArgs], optional
        Dictionary with specified arguments.
    actor_uid : Optional[str], optional
        Actor uid, by default ``None``.
    hid: Optional[str], optional
        Hid, by default ``None``.

    Returns
    -------
    scmd : Dict
        Dictionary of server command.

    """
    server_command: Dict[str, Any] = {
        "projects": [
            {"commands": [_gen_command(command=command, args=args, actor_uid=actor_uid, hid=hid)]}
        ]
    }

    if password:
        server_command["Password"] = password
    return server_command


def _gen_command(
    command: str,
    args: Optional[CommandArgs] = None,
    actor_uid: Optional[str] = None,
    hid: Optional[str] = None,
) -> Dict:
    """Generate "commands" for method server command.

    Parameters
    ----------
    command : str
        Command name.
    args: Optional[dict], optional
        Arguments, by default ``None``.
    actor_uid: Optional[str], optional
        Actor uid, by default ``None``.
    hid: Optional[str], optional
        Actor hid, by default ``None``.

    Returns
    -------
    dictionary: dict
        Specified dictionary.
    """
    cmd: Dict[str, Any] = {}
    cmd["type"] = _builtin
    cmd["command"] = command
    if args:
        cmd["args"] = args
    if actor_uid:
        cmd["actor_uid"] = actor_uid
    if hid:
        cmd["hid"] = hid
    return cmd


def _to_json(dict: Dict) -> str:
    """Convert dictionary to JSON.

    Parameters
    ----------
    dict : Dict
        Dictionary with specified query inputs.

    Returns
    -------
    str
        JSON string.
    """
    return json.dumps(dict, sort_keys=True, ensure_ascii=False)



# Copyright (C) 2022 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Module for input/output functionality."""
from __future__ import annotations

from enum import Enum
from pathlib import Path
import time
from typing import Optional, Union

from ansys.optislang.core.utils import enum_from_str


class File:
    """Provides for operating on files."""

    def __init__(self, path: Union[Path, str]) -> None:
        """Create a ``File`` instance.

        Parameters
        ----------
        path : Union[pathlib.Path, str]
            Path to the file.
        id: str, optional
            File ID.
        comment:  str, optional
            Description of the file.
        """
        self.__path = Path(path)

    def __eq__(self, other) -> bool:
        r"""Compare properties of two instances of the ``File`` class.

        Parameters
        ----------
        other: File
            File for comparison.

        Returns
        -------
        bool
            ``True`` if all properties match, ``False`` otherwise.
        """
        return (
            type(self) == type(other)
            and self.exists == other.exists
            and self.filename == other.filename
            and self.last_modified_seconds == other.last_modified_seconds
            and self.last_modified_str == other.last_modified_str
            and self.path == other.path
            and self.size == other.size
        )

    @property
    def exists(self) -> bool:
        """Check whether file exists.

        Returns
        -------
        bool
            Information, whether file exists.
        """
        return self.path.exists()

    @property
    def filename(self) -> str:
        """Get file name.

        Returns
        -------
        str
            File name
        """
        return self.path.name

    @property
    def last_modified_seconds(self) -> Optional[float]:
        """Last modified time as a timestamp.

        Returns
        -------
        Optional[float]
            Last modified time in seconds since the Epoch, `None` if file doesn't exist.
        """
        if self.exists:
            return self.path.stat().st_mtime
        else:
            return None

    @property
    def last_modified_str(self) -> Optional[str]:
        """Last modified time as a datetime.

        Returns
        -------
        Optional[str]
            Last modified time as string, `None` if file doesn't exist.
        """
        if self.exists:
            return time.ctime(self.last_modified_seconds)
        else:
            return None

    @property
    def path(self) -> Path:
        """Path to the file.

        Returns
        -------
        Path
            Path the the file
        """
        return self.__path

    @property
    def size(self) -> Optional[int]:
        """File size in bytes.

        Returns
        -------
        Optional[int]
            File size in bytes, `None` in file doesn't exist.
        """
        if self.exists:
            return self.path.stat().st_size
        else:
            return None


class RegisteredFile(File):
    """Provides for operating on registered files."""

    def __init__(
        self,
        path: Union[Path, str],
        id: str,
        comment: str,
        tag: str,
        usage: Union[RegisteredFileUsage, str],
    ) -> None:
        """Create a ``RegisteredFile`` instance.

        Parameters
        ----------
        path : Union[pathlib.Path, str]
            Path to the file.
        id: str
            File ID.
        comment:  str
            Description of the file.
        tag: str
            File uid.
        usage: Union[RegisteredFileUsage, str]
            Usage of registered file.
        """
        super().__init__(path=path)
        self.__id = id
        self.__comment = comment
        self.__tag = tag
        if not isinstance(usage, RegisteredFileUsage):
            usage = RegisteredFileUsage.from_str(usage)
        self.__usage = usage

    def __eq__(self, other) -> bool:
        r"""Compare properties of two instances of the ``RegisteredFile`` class.

        Parameters
        ----------
        other: RegisteredFile
            File for comparison.

        Returns
        -------
        bool
            ``True`` if all properties match, ``False`` otherwise.
        """
        return (
            type(self) == type(other)
            and self.comment == other.comment
            and self.id == other.id
            and self.tag == other.tag
            and self.usage == other.usage
            and super().__eq__(other)
        )

    @property
    def comment(self) -> str:
        """File description.

        Returns
        -------
        str
            File description.
        """
        return self.__comment

    @property
    def id(self) -> str:
        """File id.

        Returns
        -------
        str
            File id.
        """
        return self.__id

    @property
    def tag(self) -> str:
        """File uid.

        Returns
        -------
        str
            File uid.
        """
        return self.__tag

    @property
    def usage(self) -> RegisteredFileUsage:
        """Usage of registered file.

        Returns
        -------
        RegisteredFileUsage
            Usage of registered file.
        """
        return self.__usage


class FileOutputFormat(Enum):
    """Provides options for storing files."""

    JSON = 0
    CSV = 1

    def to_str(self) -> str:
        """Convert file type to suffix.

        Returns
        -------
        str
            File suffix.
        """
        return "." + self.name.lower()


class RegisteredFileUsage(Enum):
    """Provides options for registered files usage."""

    UNDETERMINED = 0
    INTERMEDIATE_RESULT = 1
    INPUT_FILE = 2
    OUTPUT_FILE = 3

    @classmethod
    def from_str(cls, string: str) -> RegisteredFileUsage:
        """Convert string to an instance of the ``RegisteredFileUsage`` class.

        Parameters
        ----------
        string: str
            String to be converted.

        Returns
        -------
        RegisteredFileUsage
            Instance of the ``RegisteredFileUsage`` class.

        Raises
        ------
        TypeError
            Raised when an invalid type of ``string`` is given.
        ValueError
            Raised when an invalid value of ``string`` is given.
        """
        return enum_from_str(string=string, enum_class=cls, replace=(" ", "_"))

    def to_str(self) -> str:
        """Convert usage type to string.

        Returns
        -------
        str
            Usage type.
        """
        return self.name[0] + self.name[1:].lower().replace("_", " ")



# Copyright (C) 2022 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Contains classes to obtain operate with project parametric."""
from __future__ import annotations

from typing import TYPE_CHECKING, Any, Dict, List, Optional, Tuple

from ansys.optislang.core.managers import CriteriaManager, ParameterManager, ResponseManager
from ansys.optislang.core.project_parametric import (
    ConstraintCriterion,
    Criterion,
    LimitStateCriterion,
    Parameter,
    Response,
)

if TYPE_CHECKING:
    from ansys.optislang.core.tcp.osl_server import TcpOslServer


class TcpCriteriaManagerProxy(CriteriaManager):
    """Contains methods for obtaining criteria."""

    __PROPERTY_MAPPING = {"limit_expression": "limit"}

    def __init__(self, uid: str, osl_server: TcpOslServer) -> None:
        """Initialize a new instance of the ``TcpCriteriaManagerProxy`` class.

        Parameters
        ----------
        uid: str
            Unique ID of the instance.
        osl_server: TcpOslServer
            Object providing access to the optiSLang server.
        """
        self.__uid = uid
        self.__osl_server = osl_server

    def __str__(self) -> str:
        """Get the unique ID of the criteria manager."""
        return f"CriteriaManager uid: {self.__uid}"

    def add_criterion(self, criterion: Criterion) -> None:
        """Add criterion to the system.

        Parameters
        ----------
        criterion: Criterion
            Criterion to be created in the system.

        Raises
        ------
        NotImplementedError
            Raised when unsupported optiSLang server is used.
        OslCommunicationError
            Raised when an error occurs while communicating with the server.
        OslCommandError
            Raised when a command or query fails.
        TimeoutError
            Raised when the timeout float value expires.
        """
        if (self.__osl_server.osl_version.major + self.__osl_server.osl_version.minor / 10) < 23.2:
            raise NotImplementedError("Method is supported for Ansys optiSLang version >= 23.2.")
        else:
            self.__osl_server.add_criterion(
                uid=self.__uid,
                criterion_type=criterion.criterion.name,
                expression=criterion.expression,
                name=criterion.name,
                limit=(
                    criterion.limit_expression
                    if isinstance(criterion, (ConstraintCriterion, LimitStateCriterion))
                    else None
                ),
            )

    def get_criteria(self) -> Tuple[Criterion, ...]:
        """Get the criteria of the system.

        Returns
        -------
        Tuple[Criterion, ...]
            Tuple of the criterion for the system.

        Raises
        ------
        OslCommunicationError
            Raised when an error occurs while communicating with the server.
        OslCommandError
            Raised when a command or query fails.
        TimeoutError
            Raised when the timeout float value expires.
        """
        container = (
            self.__osl_server.get_actor_properties(uid=self.__uid)
            .get("Criteria", {})
            .get("sequence", [{}])
        )
        return tuple([Criterion.from_dict(criterion_dict) for criterion_dict in container])

    def get_criteria_names(self) -> Tuple[str, ...]:
        """Get all criteria names.

        Returns
        -------
        Tuple[str, ...]
            Tuple of all criteria names.

        Raises
        ------
        OslCommunicationError
            Raised when an error occurs while communicating with the server.
        OslCommandError
            Raised when a command or query fails.
        TimeoutError
            Raised when the timeout float value expires.
        """
        container = (
            self.__osl_server.get_actor_properties(uid=self.__uid)
            .get("Criteria", {})
            .get("sequence", [{}])
        )
        return tuple([criterion_dict["First"] for criterion_dict in container])

    def modify_criterion(self, criterion: Criterion) -> None:
        """Modify criterion in the system.

        Parameters
        ----------
        criterion : Criterion
            Criterion to be modified. Criterion name is used as idefentifier.

        Raises
        ------
        NotImplementedError
            Raised when unsupported optiSLang server is used.
        OslCommunicationError
            Raised when an error occurs while communicating with the server.
        OslCommandError
            Raised when a command or query fails.
        TimeoutError
            Raised when the timeout float value expires.
        """
        if (self.__osl_server.osl_version.major + self.__osl_server.osl_version.minor / 10) < 23.2:
            raise NotImplementedError("Method is supported for optiSLang version >= 23.2.")
        else:
            self.modify_criterion_property(
                criterion_name=criterion.name,
                property_name="type",
                property_value=criterion.criterion.name,
            )
            self.modify_criterion_property(
                criterion_name=criterion.name,
                property_name="expression",
                property_value=criterion.expression,
            )
            if isinstance(criterion, (ConstraintCriterion, LimitStateCriterion)):
                self.modify_criterion_property(
                    criterion_name=criterion.name,
                    property_name="limit",
                    property_value=criterion.limit_expression,
                )

    def modify_criterion_property(
        self, criterion_name: str, property_name: str, property_value: Any
    ) -> None:
        """Modify property of criterion in the system.

        Parameters
        ----------
        criterion_name : str
            Name of criterion to be modified.
        property_name: str
            Name of property to be modified.
        property_value: Any
            New value of the modified property.

        Raises
        ------
        NotImplementedError
            Raised when unsupported optiSLang server is used.
        OslCommunicationError
            Raised when an error occurs while communicating with the server.
        OslCommandError
            Raised when a command or query fails.
        TimeoutError
            Raised when the timeout float value expires.
        """
        if (self.__osl_server.osl_version.major + self.__osl_server.osl_version.minor / 10) < 23.2:
            raise NotImplementedError("Method is supported for optiSLang version >= 23.2.")
        else:
            self.__osl_server.set_criterion_property(
                uid=self.__uid,
                criterion_name=criterion_name,
                name=self.__PROPERTY_MAPPING.get(property_name, property_name),
                value=property_value,
            )

    def remove_all_criteria(self) -> None:
        """Remove all criteria from the system.

        Raises
        ------
        NotImplementedError
            Raised when unsupported optiSLang server is used.
        OslCommunicationError
            Raised when an error occurs while communicating with the server.
        OslCommandError
            Raised when a command or query fails.
        TimeoutError
            Raised when the timeout float value expires.
        """
        if (self.__osl_server.osl_version.major + self.__osl_server.osl_version.minor / 10) < 23.2:
            raise NotImplementedError("Method is supported for optiSLang version >= 23.2.")
        else:
            self.__osl_server.remove_criteria(uid=self.__uid)

    def remove_criterion(self, criterion_name: str) -> None:
        """Remove criterion from the system.

        Parameters
        ----------
        criterion_name : str
            Name of the criterion to be removed.

        Raises
        ------
        NotImplementedError
            Raised when unsupported optiSLang server is used.
        OslCommunicationError
            Raised when an error occurs while communicating with the server.
        OslCommandError
            Raised when a command or query fails.
        TimeoutError
            Raised when the timeout float value expires.
        """
        if (self.__osl_server.osl_version.major + self.__osl_server.osl_version.minor / 10) < 23.2:
            raise NotImplementedError("Method is supported for optiSLang version >= 23.2.")
        else:
            self.__osl_server.remove_criterion(uid=self.__uid, name=criterion_name)


class TcpParameterManagerProxy(ParameterManager):
    """Contains methods for obtaining parameters."""

    __PROPERTY_MAPPING = {
        "operation": "dependency_expression",
    }

    def __init__(self, uid: str, osl_server: TcpOslServer) -> None:
        """Initialize a new instance of the ``TcpParameterManagerProxy`` class.

        Parameters
        ----------
        uid: str
            Unique ID of the instance.
        osl_server: TcpOslServer
            Object providing access to the optiSLang server.
        """
        self.__uid = uid
        self.__osl_server = osl_server

    def __str__(self) -> str:
        """Get the unique ID of the parameter manager."""
        return f"ParameterManager uid: {self.__uid}"

    def add_parameter(self, parameter: Parameter) -> None:
        """Add parameter to the system.

        Parameters
        ----------
        parameter: Parameter
            Parameter to be created in the system.

        Raises
        ------
        NameError
            Raised when the parameter with the given name already exists.
        OslCommunicationError
            Raised when an error occurs while communicating with the server.
        OslCommandError
            Raised when a command or query fails.
        TimeoutError
            Raised when the timeout float value expires.
        """
        parameter_manager, container = self.__get_parameter_container()
        if self.__get_parameter_idx(container=container, parameter_name=parameter.name) is not None:
            raise NameError(
                f"Parameter `{parameter.name}` already exists, choose another name"
                " or modify this parameter instead."
            )
        else:
            container.append(parameter.to_dict())
            parameter_manager["parameter_container"] = container
            self.__osl_server.set_actor_property(
                actor_uid=self.__uid, name="ParameterManager", value=parameter_manager
            )

    def get_parameters(self) -> Tuple[Parameter, ...]:
        """Get the parameters of the system.

        Returns
        -------
        Tuple[Parameter, ...]
            Tuple of the parameters for the system.

        Raises
        ------
        OslCommunicationError
            Raised when an error occurs while communicating with the server.
        OslCommandError
            Raised when a command or query fails.
        TimeoutError
            Raised when the timeout float value expires.
        """
        parameter_manager, container = self.__get_parameter_container()
        return tuple([Parameter.from_dict(parameter_dict) for parameter_dict in container])

    def get_parameters_names(self) -> Tuple[str, ...]:
        """Get all parameter names.

        Returns
        -------
        Tuple[str, ...]
            Tuple of all parameter names.

        Raises
        ------
        OslCommunicationError
            Raised when an error occurs while communicating with the server.
        OslCommandError
            Raised when a command or query fails.
        TimeoutError
            Raised when the timeout float value expires.
        """
        parameter_manager, container = self.__get_parameter_container()
        return tuple([parameter["name"] for parameter in container])

    def modify_parameter(self, parameter: Parameter):
        """Modify parameter in the system.

        Parameters
        ----------
        parameter: Parameter
            Parameter to be modified. Parameter name is used as identifier.

        Raises
        ------
        NameError
            Raised when the parameter with the given name doesn't exist.
        OslCommunicationError
            Raised when an error occurs while communicating with the server.
        OslCommandError
            Raised when a command or query fails.
        TimeoutError
            Raised when the timeout float value expires.
        """
        parameter_manager, container = self.__get_parameter_container()
        idx = self.__get_parameter_idx(container=container, parameter_name=parameter.name)
        if idx is not None:
            parameter.id = container[idx]["id"]
            container[idx] = parameter.to_dict()
            parameter_manager["parameter_container"] = container
            self.__osl_server.set_actor_property(
                actor_uid=self.__uid, name="ParameterManager", value=parameter_manager
            )
        else:
            raise NameError(
                f"Parameter `{parameter.name}` doesn't exist in current parametric system."
            )

    def modify_parameter_property(
        self, parameter_name: str, property_name: str, property_value: Any
    ) -> None:
        """Modify property of parameter in the system.

        Parameters
        ----------
        parameter_name: str
            Name of the parameter to be modified.
        property_name: str
            Name of the property to be modified.
        property_value: Any
            New value of the modified property.

        Raises
        ------
        NameError
            Raised when the parameter with the given name doesn't exists.
        OslCommunicationError
            Raised when an error occurs while communicating with the server.
        OslCommandError
            Raised when a command or query fails.
        TimeoutError
            Raised when the timeout float value expires.
        """
        parameter_manager, container = self.__get_parameter_container()
        idx = self.__get_parameter_idx(container=container, parameter_name=parameter_name)
        if idx is not None:
            container[idx][
                self.__PROPERTY_MAPPING.get(property_name, property_name)
            ] = property_value
            parameter_manager["parameter_container"] = container
            self.__osl_server.set_actor_property(
                actor_uid=self.__uid, name="ParameterManager", value=parameter_manager
            )
        else:
            raise NameError(
                f"Parameter `{parameter_name}` doesn't exist in current parametric system."
            )

    def remove_all_parameters(self) -> None:
        """Remove all parameters from the system.

        Raises
        ------
        OslCommunicationError
            Raised when an error occurs while communicating with the server.
        OslCommandError
            Raised when a command or query fails.
        TimeoutError
            Raised when the timeout float value expires.
        """
        parameter_manager, container = self.__get_parameter_container()
        parameter_manager["parameter_container"] = []
        self.__osl_server.set_actor_property(
            actor_uid=self.__uid, name="ParameterManager", value=parameter_manager
        )

    def remove_parameter(self, parameter_name: str) -> None:
        """Remove parameter from the system.

        Parameters
        ----------
        parameter_name : str
            Name of the parameter to be removed.

        Raises
        ------
        NotImplementedError
            Raised when unsupported optiSLang server is used.
        OslCommunicationError
            Raised when an error occurs while communicating with the server.
        OslCommandError
            Raised when a command or query fails.
        TimeoutError
            Raised when the timeout float value expires.
        """
        parameter_manager, container = self.__get_parameter_container()
        idx = self.__get_parameter_idx(container=container, parameter_name=parameter_name)
        if idx is not None:
            container.pop(idx)
            parameter_manager["parameter_container"] = container
            self.__osl_server.set_actor_property(
                actor_uid=self.__uid, name="ParameterManager", value=parameter_manager
            )

        else:
            raise NameError(
                f"Parameter `{parameter_name}` doesn't exist in current parametric system."
            )

    def __get_parameter_container(self) -> Tuple[Dict[str, list], List[dict]]:
        parameter_manager = self.__osl_server.get_actor_properties(uid=self.__uid)[
            "ParameterManager"
        ]
        container = parameter_manager["parameter_container"]
        return parameter_manager, container

    @staticmethod
    def __get_parameter_idx(container: list, parameter_name: str) -> Optional[int]:
        for idx, parameter in enumerate(container):
            if parameter["name"] == parameter_name:
                return idx
        return None


class TcpResponseManagerProxy(ResponseManager):
    """Contains methods for obtaining responses."""

    def __init__(self, uid: str, osl_server: TcpOslServer) -> None:
        """Initialize a new instance of the ``TcpResponseManagerProxy`` class.

        Parameters
        ----------
        uid: str
            Unique ID of the instance.
        osl_server: TcpOslServer
            Object providing access to the optiSLang server.
        """
        self.__uid = uid
        self.__osl_server = osl_server

    def __str__(self) -> str:
        """Get the unique ID of the response manager."""
        return f"ResponseManager uid: {self.__uid}"

    def get_responses(self) -> Tuple[Response, ...]:
        """Get the responses of the system.

        Returns
        -------
        Tuple[Criterion, ...]
            Tuple of the responses for the system.

        Raises
        ------
        OslCommunicationError
            Raised when an error occurs while communicating with the server.
        OslCommandError
            Raised when a command or query fails.
        TimeoutError
            Raised when the timeout float value expires.
        """
        info = self.__osl_server.get_actor_info(uid=self.__uid)
        container = info.get("responses", {})
        responses = []
        for key, res_dict in container.items():
            responses.append(Response.from_dict(key, res_dict))
        return tuple(responses)

    def get_responses_names(self) -> Tuple[str, ...]:
        """Get all responses names.

        Returns
        -------
        Tuple[str, ...]
            Tuple of all responses names.

        Raises
        ------
        OslCommunicationError
            Raised when an error occurs while communicating with the server.
        OslCommandError
            Raised when a command or query fails.
        TimeoutError
            Raised when the timeout float value expires.
        """
        info = self.__osl_server.get_actor_info(uid=self.__uid)
        container = info.get("responses", {})
        return tuple(container.keys())



# Copyright (C) 2022 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Module that returns paths to examples."""

import os
from pathlib import Path
from typing import Dict, Optional, Tuple

ExampleFiles = Tuple[Path, ...]

module_path = Path(__file__).parent.absolute()

# dictionary of files, that must be available to run scripts
example_files: Dict[str, Optional[ExampleFiles]] = {
    "ansys_workbench_portscan": None,
    "arsm_ten_bar_truss": (
        module_path / "00_run_script" / "ten_bar_truss" / "files" / "ten_bar_truss.s",
        module_path / "00_run_script" / "ten_bar_truss" / "files" / "ten_bar_truss.out",
    ),
    "ansys_workbench_ten_bar_truss": (
        module_path / "00_run_script" / "ten_bar_truss" / "files" / "ten_bar_truss_apdl.wbpz",
    ),
    "ten_bar_modify_parameters": None,
    "ten_bar_truss_lc2": (
        module_path / "00_run_script" / "ten_bar_truss" / "files" / "ten_bar_truss2.s",
        module_path / "00_run_script" / "ten_bar_truss" / "files" / "ten_bar_truss2.out",
    ),
    "oscillatorcalibration_system_ascii": (
        module_path / "00_run_script" / "oscillator" / "scripts" / "_create_oscillator.py",
        module_path / "00_run_script" / "oscillator" / "files" / "oscillator.s",
        module_path / "00_run_script" / "oscillator" / "files" / "oscillator.bat",
        module_path / "00_run_script" / "oscillator" / "files" / "oscillator.sh",
        module_path / "00_run_script" / "oscillator" / "files" / "oscillator_signal.txt",
        module_path / "00_run_script" / "oscillator" / "files" / "oscillator_reference.txt",
    ),
    "oscillatorcalibration_system_python": (
        module_path / "00_run_script" / "oscillator" / "scripts" / "_create_oscillator.py",
        module_path / "00_run_script" / "oscillator" / "files" / "oscillator_reference.txt",
    ),
    "oscillator_optimization_ea": (
        module_path / "00_run_script" / "oscillator" / "scripts" / "_create_oscillator.py",
    ),
    "oscillator_optimization_on_mop": (
        module_path / "00_run_script" / "oscillator" / "scripts" / "_create_oscillator.py",
    ),
    "oscillator_robustness_arsm": (
        module_path / "00_run_script" / "oscillator" / "scripts" / "_create_oscillator.py",
    ),
    "oscillator_sensitivity_mop": (
        module_path / "00_run_script" / "oscillator" / "scripts" / "_create_oscillator.py",
    ),
    "oscillator_system_python": (
        module_path / "00_run_script" / "oscillator" / "scripts" / "_create_oscillator.py",
    ),
    "create_all_possible_nodes": None,
    "etk_abaqus": (
        module_path / "00_run_script" / "etk_abaqus" / "files" / "oscillator.inp",
        module_path / "00_run_script" / "etk_abaqus" / "files" / "oscillator.odb",
    ),
    "python_help": None,
    "python_node": None,
    "optimizer_settings": None,
    "sensitivity_settings": None,
    "simple_calculator": (module_path / "00_run_script" / "files" / "calculator.opf",),
    "calculator_with_params": (
        module_path / "00_run_script" / "files" / "calculator_with_params.opf",
    ),
    "nodes_connection": (module_path / "00_run_script" / "files" / "connect_nodes.opf",),
    "nested_systems": (module_path / "00_run_script" / "files" / "nested_systems.opf",),
    "ten_bar_truss": (module_path / "00_run_script" / "files" / "ten_bar_truss.opf",),
    "omdb_files": (module_path / "00_run_script" / "files" / "omdb_files.opf",),
}

# dictionary of scripts to be run
example_scripts: Dict[str, Optional[Path]] = {
    "ansys_workbench_portscan": module_path
    / "00_run_script"
    / "scripts"
    / "ansys_workbench_portscan.py",
    "arsm_ten_bar_truss": module_path
    / "00_run_script"
    / "ten_bar_truss"
    / "scripts"
    / "arsm_ten_bar_truss.py",
    "ansys_workbench_ten_bar_truss": module_path
    / "00_run_script"
    / "ten_bar_truss"
    / "scripts"
    / "ansys_workbench_ten_bar_truss.py",
    "ten_bar_modify_parameters": module_path
    / "00_run_script"
    / "ten_bar_truss"
    / "scripts"
    / "ten_bar_modify_parameters.py",
    "ten_bar_truss_lc2": module_path
    / "00_run_script"
    / "ten_bar_truss"
    / "scripts"
    / "ten_bar_truss_lc2.py",
    "oscillatorcalibration_system_ascii": module_path
    / "00_run_script"
    / "oscillator"
    / "scripts"
    / "oscillatorcalibration_system_ascii.py",
    "oscillatorcalibration_system_python": module_path
    / "00_run_script"
    / "oscillator"
    / "scripts"
    / "oscillatorcalibration_system_python.py",
    "oscillator_optimization_ea": module_path
    / "00_run_script"
    / "oscillator"
    / "scripts"
    / "oscillator_optimization_ea.py",
    "oscillator_optimization_on_mop": module_path
    / "00_run_script"
    / "oscillator"
    / "scripts"
    / "oscillator_optimization_on_mop.py",
    "oscillator_robustness_arsm": module_path
    / "00_run_script"
    / "oscillator"
    / "scripts"
    / "oscillator_robustness_arsm.py",
    "oscillator_sensitivity_mop": module_path
    / "00_run_script"
    / "oscillator"
    / "scripts"
    / "oscillator_sensitivity_mop.py",
    "oscillator_system_python": module_path
    / "00_run_script"
    / "oscillator"
    / "scripts"
    / "oscillator_system_python.py",
    "create_all_possible_nodes": module_path
    / "00_run_script"
    / "scripts"
    / "create_all_possible_nodes.py",
    "etk_abaqus": module_path / "00_run_script" / "etk_abaqus" / "scripts" / "etk_abaqus.py",
    "python_help": module_path / "00_run_script" / "python" / "scripts" / "python_help.py",
    "python_node": module_path / "00_run_script" / "python" / "scripts" / "python_node.py",
    "optimizer_settings": module_path / "00_run_script" / "scripts" / "optimizer_settings.py",
    "sensitivity_settings": module_path / "00_run_script" / "scripts" / "sensitivity_settings.py",
    "simple_calculator": module_path / "00_run_script" / "scripts" / "simple_calculator.py",
    "calculator_with_params": None,
    "nodes_connection": None,
    "nested_systems": None,
    "ten_bar_truss": None,
    "omdb_files": None,
}



# Copyright (C) 2022 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Contains utility functions for encoding and decoding text."""

import sys

from ansys.optislang.core import utils

defenc = sys.getdefaultencoding()

if sys.version_info[0] >= 3:
    text_type = str
    binary_type = bytes
else:
    if defenc == "ascii":
        defenc = "utf-8"
    text_type = unicode
    binary_type = str

string_types = (binary_type,)


def safe_decode(s):
    """Safely decodes a binary_type to text_type."""
    if isinstance(s, text_type):
        return s
    elif isinstance(s, string_types):
        return s.decode(defenc, "surrogateescape")
    elif s is not None:
        raise TypeError("Expected bytes or text, but got %r" % (s,))


def to_ascii_safe(text: str) -> str:
    """Safely converts text encoding to ASCII standard.

    Conversion is performed ignoring the unencodable unicode characters.

    Parameters
    ----------
    text : str
        String which is supposed to be encoded to ASCII standard.

    Returns
    -------
    str
        String encoded in ASCII standard.
    """
    return text.encode("ascii", "ignore").decode("ascii")


def force_bytes(text: str, encoding="ascii") -> bytes:
    """Encode data to bytes ignoring all errors.

    Special characters are ignored.

    Parameters
    ----------
    text : str
        String which is supposed to be encoded.
    encoding :
        String encoding.

    Returns
    -------
    bytes
        An encoded version of the string as a bytes object.
    """
    if utils.is_iron_python():
        return binary_type(text.encode(encoding, "ignore"), encoding, "ignore")  # type: ignore
    else:
        return text.encode(encoding, "ignore")


def force_text(data: bytes, encoding="utf-8") -> str:
    """Decode bytes to text ignoring all errors.

    Special characters are ignored.

    Parameters
    ----------
    data : bytes
        Bytes which is supposed to be decoded.
    encoding :
        The encoding with which to decode the bytes.

    Returns
    -------
    str
        Decoded string.
    """
    if isinstance(data, string_types):
        return data.decode(encoding, errors="ignore")

    if sys.version_info[0] >= 3:
        return text_type(data, encoding)
    else:
        return text_type(data)



# Copyright (C) 2022 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Contains abstract optiSLang server class."""
from __future__ import annotations

from abc import ABC, abstractmethod
from typing import NamedTuple, Optional

from ansys.optislang.core import utils


class OslVersion(NamedTuple):
    """optiSLang version.

    Attributes:
        major: int
            The major version number.
        minor: int
            The minor version number.
        maintenance: int
            The maintenance version number.
        revision: int
            The revision number.
    """

    major: int
    minor: int
    maintenance: int
    revision: int


class OslServer(ABC):
    """Base class for classes which provide access to optiSLang server."""

    @abstractmethod
    def __init__(self):  # pragma: no cover
        """``OslServer`` class is an abstract base class and cannot be instantiated."""
        pass

    @property
    def is_remote(self) -> bool:
        """Determines whether the optiSLang server does not run on localhost."""
        return self.host is not None and self.port is not None and not utils.is_localhost(self.host)

    @property
    @abstractmethod
    def host(self) -> Optional[str]:  # pragma no cover
        """Get optiSLang server address or domain name.

        Get a string representation of an IPv4/v6 address or domain name
        of the running optiSLang server.

        Returns
        -------
        host: Optional[int]
            The IPv4/v6 address or domain name of the running optiSLang server, if applicable.
            Defaults to ``None``.
        """
        pass

    @property
    @abstractmethod
    def osl_version(
        self,
    ) -> OslVersion:  # pragma: no cover
        """Version of used optiSLang.

        Returns
        -------
        OslVersion
            optiSLang version as typing.NamedTuple containing
            major, minor, maintenance and revision versions.
        """
        pass

    @property
    @abstractmethod
    def osl_version_string(self) -> str:  # pragma: no cover
        """Version of used optiSLang.

        Returns
        -------
        str
            optiSLang version.
        """
        pass

    @property
    @abstractmethod
    def port(self) -> Optional[int]:  # pragma: no cover
        """Get the port the osl server is listening on.

        Returns
        -------
        port: Optional[int]
            The port the osl server is listening on, if applicable.
            Defaults to ``None``.
        """
        pass

    @property
    @abstractmethod
    def timeout(self) -> Optional[float]:  # pragma: no cover
        """Get the timeout value for executing commands.

        Returns
        -------
        timeout: Optional[float]
            Timeout in seconds to perform commands. This value must be greater
            than zero or ``None``. The default is ``None``. Another function
            raises a timeout exception if the timeout value has elapsed before
            an operation has completed. If the timeout is ``None``, functions
            wait until they're finished, and no timeout exception is raised.
        """
        pass

    @timeout.setter
    @abstractmethod
    def timeout(self, timeout: Optional[float] = 30) -> None:  # pragma: no cover
        """Set default timeout value for execution of commands.

        Parameters
        ----------
        timeout: Optional[float]
            Timeout in seconds to perform commands, it must be greater than zero or ``None``.
            Certain functions will raise a timeout exception if the timeout period value has
            elapsed before the operation has completed.
            If ``None`` is given, functions will wait until they're finished (no timeout
            exception is raised). Defaults to ``30``.
        """
        pass

    @abstractmethod
    def dispose(self) -> None:  # pragma: no cover
        """Terminate all local threads and unregister listeners.

        Raises
        ------
        OslCommunicationError
            Raised when an error occurs while communicating with server.
        OslCommandError
            Raised when the command or query fails.
        TimeoutError
            Raised when the timeout float value expires.
        """
        pass

    @abstractmethod
    def shutdown(self, force: bool = False) -> None:  # pragma: no cover
        """Shutdown the optiSLang server.

        Stop listening for incoming connections, discard pending requests, and shut down
        the server. Batch mode exclusive: Continue project run until execution finished.
        Terminate optiSLang.

        Parameters
        ----------
        force : bool, optional
            Determines whether to force shutdown the local optiSLang server. Has no effect when
            the connection is established to the remote optiSLang server. In all cases, it is tried
            to shutdown the optiSLang server process in a proper way. However, if the force
            parameter is ``True``, after a while, the process is forced to terminate and no
            exception is raised. Defaults to ``False``.

        Raises
        ------
        OslCommunicationError
            Raised when the parameter force is ``False`` and an error occurs while communicating
            with server.
        OslCommandError
            Raised when the parameter force is ``False`` and the command or query fails.
        TimeoutError
            Raised when the parameter force is ``False`` and the timeout float value expires.
        """
        pass



# Copyright (C) 2022 - 2024 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
DYNARDO example.

Damped oscillator

Undamped eigenfrequency: omega_0 = sqrt(k/m)
Damped eigenfrequency:     omega = sqrt(1-D^2)*omega_0
Response:                   x(t) = exp(-D*omega_0*t)*sqrt(2*Ekin/m)/omega*sin(omega*t)

Minimize maximum amplitude after 5 seconds
Maximum eigenfrequencies as constraint: omega <= 8 1/s

0.1 kg <= m <= 5.0 kg
10 N/m <= k <= 50 N/m
D = 0.02
Ekin = 10 Nm

For background information see ../doc/oszillator.pdf
"""
from math import *

# Check if Python is called outside optiSLang
if not "OSL_REGULAR_EXECUTION" in locals():
    OSL_REGULAR_EXECUTION = False

# values to default if not existing yet
if not OSL_REGULAR_EXECUTION:  # test run mode
    m = 1.00000000
    k = 20.00000000
    D = 0.02000000
    Ekin = 10.00000000

# scalar values
v0 = sqrt(2 * Ekin / m)
omega0 = sqrt(k / m)
omega_damped = omega0 * sqrt(1 - D * D)

# approximate maximum amplitude from envelope
x_max_env = exp(-5 * D * omega0) * v0 / omega_damped
numsteps = 100


# time signal
def CalcDisplacement(_t):
    """Calculate displacement."""
    sin_value = sin(_t * omega_damped)
    envelope = exp(-_t * D * omega0) * v0 / omega_damped
    return sin_value * envelope


def CalcIndicator(_t, _x):
    """Calculate indicator."""
    if _t > 5:
        return abs(_x)
    else:
        return 0


time = range(0, numsteps)
time = [10.0 * t / (numsteps - 1) for t in time]
x_values = [CalcDisplacement(t) for t in time]


# extract maximum amplitude from time signal
x_max = max([CalcIndicator(t, x) for t, x in zip(time, x_values)])

print("omega_damped: " + str(omega_damped))
print("x_max: " + str(x_max))

try:
    from pyvariant import List2VariantVector

    time = List2VariantVector(time)
    x_values = List2VariantVector(x_values)
except:
    print("Did not export result to optiSLang")
